# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
from __future__ import print_function
from builtins import range
import numpy as np
from cubical.flagging import FL
from cubical.machines.abstract_machine import MasterMachine
import cubical.kernels
from cubical.solver import log
import logging
import re
from numpy.ma import masked_array

def copy_or_identity(array, time_ind=0, out=None):
    if out is None:
        return array
    np.copyto(out, array)
    return out


class PerIntervalGains(MasterMachine):
    """
    This is a base class for all gain solution machines that use solutions intervals.
    """

    def __init__(self, label, data_arr, ndir, nmod, times, frequencies, chunk_label, options):
        """
        Initialises a gain machine which supports solution intervals.
        
        Args:
            label (str):
                Label identifying the Jones term.
            data_arr (np.ndarray): 
                Shape (n_mod, n_tim, n_fre, n_ant, n_ant, n_cor, n_cor) array containing observed 
                visibilities. 
            ndir (int):
                Number of directions.
            nmod (nmod):
                Number of models.
            times (np.ndarray):
                Times for the data being processed.
            freqs (np.ndarray):
                Frequencies for the data being processsed.
            options (dict): 
                Dictionary of options.
            diag_gains (bool):
                If True, gains are diagonal-only. Else gains are full 2x2.
        """

        MasterMachine.__init__(self, label, data_arr, ndir, nmod, times, frequencies,
                               chunk_label, options)

        # === Kernel selection
        # We have three types of (progressively faster) kernels: 2x2, Diag (diag gains), and DiagDiag (diag gains and data).
        # We need to choose a kernel for the solve process, and a kernel for computing full residuals.
        # The choice is based on three binary options that control kernel selection:
        #
        #  1. diag-data   (True if data is null on the off-diagonals)
        #  2. diag-only   (True if only the diagonal terms should be used for the solution)
        #  3. is_diagonal (True if the gain matrix is diagonal)
        #
        # Here's the selection matrix:
        #
        #  diag-data  diag-only  is_diagonal    solve kernel    full kernel
        #  ---------  ---------  -----------    ------------    -----------
        #      0          0           0             2x2            2x2
        #      0          0           1            diag           diag
        #      0          1           0             2x2            2x2
        #      0          1           1          diag-diag        diag
        #      1          0           0             2x2            2x2
        #      1          0           1          diag-diag      diag-diag
        #      1          1           0             2x2            2x2
        #      1          1           1          diag-diag      diag-diag

        # select which kernels to use for computing full data
        self.kernel = self.get_full_kernel(options, self.is_diagonal)

        # as per matrix above, the solve kernel is the same as the full kernel, except if
        # solving for diagonal terms only, with a diagonal gain, in which case we can use diag-diag.
        if not options.get('diag-data') and self.diag_only and self.is_diagonal:
            self.kernel_solve = cubical.kernels.import_kernel('diagdiag_complex')
        else:
            self.kernel_solve = self.kernel

        log(2).print("{} kernels are {} {}".format(label, self.kernel, self.kernel_solve))

        self.t_int = options["time-int"] or self.n_tim
        self.f_int = options["freq-int"] or self.n_fre
        self.eps = 1e-6

        # Initialise attributes used for computing values over intervals.
        # n_tim and n_fre are the time and frequency dimensions of the data arrays.
        # n_timint and n_freint are the time and frequency dimensions of the gains.

        self.t_bins = list(range(0, self.n_tim, self.t_int))
        self.f_bins = list(range(0, self.n_fre, self.f_int))

        self.n_timint = len(self.t_bins)
        self.n_freint = len(self.f_bins)
        self.n_tf_ints = self.n_timint * self.n_freint

        # number of valid solutions
        self.n_valid_sols = self.n_dir * self.n_tf_ints

        # split grids into intervals, and find the centre of gravity of each
        timebins = np.split(times, self.t_bins[1:])
        freqbins = np.split(frequencies, self.f_bins[1:])
        timegrid = np.array([float(x.mean()) for x in timebins])
        freqgrid = np.array([float(x.mean()) for x in freqbins])

        # interval_grid determines the per-interval grid poins
        self.interval_grid = dict(time=timegrid, freq=freqgrid)
        # data_grid determines the full resolution grid
        self.data_grid = dict(time=times, freq=frequencies)

        # compute index from each data point to interval number
        t_ind = np.arange(self.n_tim)//self.t_int
        f_ind = np.arange(self.n_fre)//self.f_int

        self.t_mapping, self.f_mapping = np.meshgrid(t_ind, f_ind, indexing="ij")

        # Initialise attributes used in convergence testing. n_cnvgd is the number
        # of solutions which have converged.

        self._has_stalled = False
        self.n_cnvgd = 0
        self._frac_cnvgd = 0
        self.iters = 0
        self.min_quorum = options["conv-quorum"]
        self.update_type = set(options["update-type"].split("-"))
        self.ref_ant = options["ref-ant"]
        if self.ref_ant is not None:
            if type(self.ref_ant) is not str:
                self.ref_ant = str(self.ref_ant)
            if self.ref_ant[0] == "#":
                self.ref_ant = int(self.ref_ant[1:])
            else:
                from cubical.solver import metadata
                self.ref_ant = metadata.antenna_index[self.ref_ant]

        self.fix_directions = options["fix-dirs"] if options["fix-dirs"] is not None and \
                options["fix-dirs"] != "" else []

        if type(self.fix_directions) is int:
            self.fix_directions = [self.fix_directions]
        if type(self.fix_directions) is str and re.match(r"^\W*\d{1,}(\W*,\W*\d{1,})*\W*$", self.fix_directions):
            self.fix_directions = map(int, map(str.strip, ",".split(self.fix_directions)))

        if not (type(self.fix_directions) is list and
                all(map(lambda x: type(x) is int, self.fix_directions))):
            raise ValueError("fix-dir must be number or list of numbers")

        # True if gains are loaded from a DB
        self._gains_loaded = False

        # Construct flag array and populate flagging attributes.
        self.max_gain_error = options["max-prior-error"]
        self.max_post_error = options["max-post-error"]
        self.low_snr_warn = options["low-snr-warn"]
        self.high_gain_var_warn = options["high-gain-var-warn"]
        self.clip_lower = options["clip-low"]
        self.clip_upper = options["clip-high"]
        self.clip_after = options["clip-after"]

        self.init_gains()
        self.old_gains = self.gains.copy()

        # mask of intervals/antennas for which we have no data (i.e. all flagged to begin with). Shape TFA.
        self.missing_intervals_ant = None
        # mask of intervals/antennas for which we have valid solutions. Note that this is not necessarily
        # a full complement to missing_intervals_ant -- some intervals/antennas may become flagged due to low SNR,
        # ill-conditioning, etc., in which case they're neither valid, nor missing. Shape DTFA.
        self.valid_intervals_ant = None

        # number of gains flagged on various conditions
        self.num_flagged_on = {}

        # Gain error estimates. Populated by subclasses, if available
        # Should be array of same shape as the gains
        self.prior_gain_error = None
        self.posterior_gain_error = None

        # buffers for arrays used in internal updates
        self._jh = self._jhr = self._jhj = self._gh = self._r = self._ginv = self._ghinv = None
        self._update = None

        # flag: have gains been updated
        self._gh_update = self._ghinv_update = True

    @classmethod
    def determine_allocators(cls, options):
        kernel = cls.get_full_kernel(options, diag_gains=cls.determine_diagonality(options))
        return kernel.allocate_vis_array, kernel.allocate_flag_array, kernel.allocate_gain_array

    @classmethod
    def get_full_kernel(cls, options, diag_gains):
        """
        Class method. Given a machine class and a set of options, returns numerical kernel
        for "full 2x2" operations on visibilities.
        """
        # return cubical.kernels.import_kernel('full_complex')
        # select which kernels to use for full correction/residuals -- see selection matrix above
        if diag_gains:
            if options.get('diag-data'):
                return cubical.kernels.import_kernel('diagdiag_complex')
            else:
                return cubical.kernels.import_kernel('diag_complex')
        else:
            return cubical.kernels.import_kernel('full_complex')

    def get_conj_gains(self):
        if self._gh is None:
            self._gh = np.empty_like(self.gains)
        if self._gh_update:
            np.conj(self.gains.transpose(0, 1, 2, 3, 5, 4), out=self._gh)
            self._gh_update = False
        return self._gh

    def get_inverse_gains(self):
        if self._ginv is None:
            self._ginv = np.empty_like(self.gains)
            self._ghinv = np.empty_like(self.gains)
        if self._ghinv_update:
            self._ghinv_flag_count = self.kernel_solve.invert_gains(
                self.gains, self._ginv, self.gflags, self.eps, FL.ILLCOND)
            np.conj(self._ginv.transpose(0, 1, 2, 3, 5, 4), out=self._ghinv)
            self._ghinv_update = False
        return self._ginv, self._ghinv, self._ghinv_flag_count

    def get_new_jh(self, model_arr):
        if self._jh is None:
            self._jh = np.empty_like(model_arr)
        # else:
        #     self._jh.fill(0)
        return self._jh

    def get_new_jhr(self):
        if self._jhr is None:
            self._jhr = np.zeros_like(self.gains)
        else:
            self._jhr.fill(0)
        return self._jhr

    def get_new_jhj(self):
        if self._jhj is None:
            self._jhj = np.zeros_like(self.gains)
            self._jhjinv = np.empty_like(self.gains)
        else:
            self._jhj.fill(0)
        return self._jhj, self._jhjinv

    def init_update(self, jhr):
        if self._update is None:
            self._update = np.zeros_like(jhr)
        else:
            self._update.fill(0)
        return self._update

    def get_obs_or_res(self, obser_arr, model_arr):
        if self.n_dir > 1:
            if self._r is None:
                self._r = np.empty_like(obser_arr)
            self.compute_residual(obser_arr, model_arr, self._r, require_full=False)
            return self._r
        else:
            return obser_arr

    def get_gain_mask(self):
        """Helper method: make a gain array mask from gain flags by broadcasting the corr1/2 axes."""
        mask = np.zeros_like(self.gains, bool)
        mask[:] = (self.gflags!=0)[...,np.newaxis,np.newaxis]
        return mask

    def init_gains(self):
        """
        Construct gain and flag arrays. Normally we have one gain/one flag per interval, but 
        subclasses may redefine this, if they deal with e.g. full-resolution gains.
        
        Sets up the following attributes: 
            gain_shape (list):
                shape of the gains array, i.e. (n_dir, NT, NF, n_ant, n_cor, n_cor)
                Default version has NT=n_timint, NF=n_freint, i.e. one gain per interval.
                 
            gain_grid (dict):
                grid on which gains are defined, as a dict of {'time': times, 'freq': frequencies},
                where times is a vector of NT points and frequencies is a vector of NF points.
                
            gains (np.ndarray):
                gains array of the specified shape
                
            gflags (np.ndarray):
                gain flags array of the same shape, minus the two correlation axes
                
            gain_intervals (list):
                intervals on which gains are defined. Default version has [t_int, f_int].
                
            _gainres_to_fullres (callable):
                function to go from an array of gain shape to full time/freq resolution
                (default uses unpack_intervals)
            
            _interval_to_gainres (callable):
                function to go from an array of interval shape to gain shape (default is identity)
        """
        # Construct the appropriate shape for the gains.
        self.gain_intervals = self.t_int, self.f_int
        self.gain_shape = [self.n_dir, self.n_timint, self.n_freint, self.n_ant, self.n_cor, self.n_cor]
        self.gain_grid = self.interval_grid

        self.gains = self.kernel_solve.allocate_gain_array(self.gain_shape, self.dtype)

        self.gains[:] = np.eye(self.n_cor)
        self.gflags = np.zeros(self.gain_shape[:-2], FL.dtype)

        # function used to unpack the gains or flags into full time/freq resolution
        self._gainres_to_fullres  = self.unpack_intervals
        # function used to unpack interval resolution to gain resolution
        self._interval_to_gainres = self.copy_or_identity

    def compute_residual(self, obser_arr, model_arr, resid_arr, require_full=True):
        gains_h = self.get_conj_gains()

        np.copyto(resid_arr, obser_arr)

        (self.kernel if require_full else self.kernel_solve).compute_residual(model_arr,
                                                                self.gains, gains_h, resid_arr, *self.gain_intervals)

        return resid_arr

    def apply_gains(self, model_arr, full2x2=True, dd_only=False):
        gains_h = self.get_conj_gains()

        (self.kernel if full2x2 else self.kernel_solve).apply_gains(model_arr,
                                        self.gains, gains_h, *self.gain_intervals)

        return model_arr

    def apply_inv_gains(self, obser_arr, corr_vis=None, full2x2=True, direction=None):
        if corr_vis is None:
            corr_vis = np.empty_like(obser_arr)
    
        # parset uses -1 for None, so may as well support it here        
        if direction is not None and direction < 0:
            direction = None

        if self.dd_term and direction is None:
            corr_vis[:] = obser_arr
            return corr_vis, 0

        g_inv, gh_inv, flag_count = self.get_inverse_gains()

        dirslice = slice(0,1) if not self.dd_term else slice(direction, direction+1)
        g_inv = g_inv[dirslice]
        gh_inv = gh_inv[dirslice]

        (self.kernel if full2x2 else self.kernel_solve).compute_corrected(obser_arr,
                                                            g_inv, gh_inv, corr_vis, *self.gain_intervals)

        return corr_vis, flag_count

    @staticmethod
    def exportable_solutions():
        """ Returns a dictionary of exportable solutions for this machine type. """

        return { "gain": (1+0j, ("dir", "time", "freq", "ant", "corr1", "corr2")),
                 "gain.err": (0., ("dir", "time", "freq", "ant", "corr1", "corr2")) }

    def importable_solutions(self, grid0):
        """ Returns a dictionary of importable solutions for this machine type. """
        return { "gain": dict(dir=grid0['dir'] if self.dd_term else [0], ant=grid0['ant'], corr1=grid0['corr'], corr2=grid0['corr'],
                              **self.interval_grid) }

    def export_solutions(self):
        """ Saves the solutions to a dict of {label: solutions,grids} items. """
        
        mask = self.get_gain_mask()

        sols = { "gain": (masked_array(self.gains, mask), self.gain_grid) }

        if self.posterior_gain_error is not None:
            sols["gain.err"] = (masked_array(self.posterior_gain_error, mask), self.gain_grid)
        else:
            sols["gain.err"] = (masked_array(np.zeros(self.gain_shape, float), np.ones(self.gain_shape, bool)),
                                self.gain_grid)

        return sols

    def import_solutions(self, soldict):
        """ 
        Loads solutions from a dict. 
        
        Args:
            soldict (dict):
                Contains gains solutions which must be loaded.
        """

        sol = soldict.get("gain")
        if sol is not None:
            self.gains[:] = sol.data
            # gain of all-1 are actually missing from the DB -- replace by unity
            all_ones = (self.gains == 1.+0j).all(axis=(-1, -2)) 
            sol.mask[all_ones,:,:] = True
            self.gains[all_ones, 0, 1] = self.gains[all_ones, 1, 0] = 0
            # flag them as NOSOL if we're not solvable
            if not self.solvable:
                self.gflags[sol.mask.any(axis=(-1,-2))] |= FL.NOSOL
            self._gains_loaded = True

        self.restrict_solution(self.gains)

    def precompute_attributes(self, data_arr, model_arr, flags_arr, inv_var_chan):
        """Precomputes various attributes of the machine before starting a solution"""
        unflagged = MasterMachine.precompute_attributes(self, data_arr, model_arr, flags_arr, inv_var_chan)

        self.missing_gain_fraction = self.missing_intervals_ant.sum() / float(self.missing_intervals_ant.size)

        # number of valid records per time/frequency/antenna (summed over baseline)
        numvis_tfa = unflagged.sum(axis=-1)

        missing_slots = numvis_tfa == 0

        missing_intervals = self.missing_intervals_ant.all(axis=-1)

        if self.solvable and missing_intervals.any() and log.verbosity() > 1:
            self.raise_userwarning(
                logging.WARNING,    # OMS: was CRITICAL but I disagree. Chunks of band/scans are often flagged, why cry wolf about it?
                "{0:s} {1:s}: {2:d}/{3:d} solution intervals are fully flagged".format(
                    self.chunk_label, self.jones_label,
                    missing_intervals.sum(), missing_intervals.size
                ),
                90, raise_once="prior_fully_flagged_dirs", verbosity=2)

        # compute error estimates per direction, antenna, and interval
        if inv_var_chan is not None:
            # collapse direction axis, if not directional
            if not self.dd_term:
                model_arr = model_arr.sum(axis=0, keepdims=True)
            # mean |model|^2 per direction+TFA, normalized by the number of equations/2
            # (a full 2x2 coherency has 8 equations, therefore 4 model components.)
            # TODO: think how this plays out in diag-only scenarios
            with np.errstate(invalid='ignore', divide='ignore'):
                modelsq = (model_arr*np.conj(model_arr)).real.sum(axis=(1,-1,-2,-3)) / \
                      (self.eqs_per_record/2 * numvis_tfa)
                sigmasq = 1.0/inv_var_chan                        # squared noise per channel. Could be infinite if no data
            modelsq[:, missing_slots] = 0
            # take the sigma (in quadrature) over each interval
            # divided by quadrature unflagged contributing interferometers per interval
            # this yields var<g>
            # (numeq_tfa becomes number of unflagged points per interval, antenna)
            numvis_int_ant = self.interval_sum(numvis_tfa)
            sigmasq[np.isnan(sigmasq)|np.isinf(sigmasq)] = 0.0
            modelsq[np.isnan(modelsq)|np.isinf(modelsq)] = 0.0

            # sum-square-model over each interval
            modelsq_sum = self.interval_sum(modelsq, 1)
            # mask of intervals where the model sum is zero or invalid
            invalid_models = (modelsq_sum == 0) | np.isnan(modelsq_sum) | np.isinf(modelsq_sum)

            # if interval was missing (fully flagged), it's not invalid (these will be masked out separately)
            invalid_models[:, self.missing_intervals_ant] = False
            if invalid_models.any():
                self.raise_userwarning(
                    logging.CRITICAL,
                    "{0:s} {1:s}: {2:d}/{3:d} directions/intervals has an invalid or null model. " +
                    "This can be caused by bad data, but should have been handled gracefully, so please report this issue!".format(
                        self.chunk_label, self.jones_label, invalid_models.sum(), invalid_models.size
                    ),
                    90, raise_once="invalid_models", verbosity=2, color="red")

            with np.errstate(invalid='ignore', divide='ignore'):
                NSR_int = self.interval_sum(np.ones_like(modelsq) * (sigmasq)[None, None, :, None], 1) / \
                              (modelsq_sum * numvis_int_ant)
                self.prior_gain_error = np.sqrt(NSR_int)

            # convert that into a gain error per direction,interval,antenna
            # zero the PGEs corresponding to fully-flagged intervals, since we don't want to shout about them
            self.prior_gain_error[:, self.missing_intervals_ant] = 0
            # zero the PGEs corresponding to invalid models, we've already warned about those
            self.prior_gain_error[invalid_models] = 0
            # zero the PGEs for fixed directions
            if self.dd_term:
                self.prior_gain_error[self.fix_directions, ...] = 0

            # if we did everything right above (ignoring flagged data etc.), PGE can't be inf/nan,
            # so this is just a sanity check
            pge_flag_invalid = np.isnan(self.prior_gain_error) | np.isinf(self.prior_gain_error)
            if pge_flag_invalid.any():
                self.raise_userwarning(
                    logging.CRITICAL,
                    "{0:s} {1:s}: prior gain error estimate invalid for {2:d}/{3:d} directions/intervals. " +
                    "This can be caused by bad data, but should have been handled gracefully, so please report this issue!".format(
                        self.chunk_label, self.jones_label, pge_flag_invalid.sum(), pge_flag_invalid.size
                    ),
                    90, raise_once="invalid_models", verbosity=2, color="red")
                # flag these as invalid
                self.gflags[self._interval_to_gainres(pge_flag_invalid, 1)] |= FL.INVALID

            # log(0).print("{}: max gain error {}, invalid PGEs {}".format(self.chunk_label, self.max_gain_error, bad_gain_intervals.sum()))

            if self.max_gain_error:
                low_snr = self.prior_gain_error > self.max_gain_error  # shape DTFA
                self.gflags[self._interval_to_gainres(low_snr, 1)] |= FL.LOWSNR
                self._report_gain_flags(low_snr, "on low SNR", "your max-prior-error settings", self.low_snr_warn)

        # if INVALID or LOWSNR raised above, propagate and recompute equation counts
        # This also propagates the NOSOL flag from missing solutions
        if self._update_gain_flags("INVALID", flags_arr) + \
                self._update_gain_flags("LOWSNR", flags_arr) + \
                self._update_gain_flags("NOSOL", flags_arr):
            unflagged = flags_arr == 0
            self.update_equation_counts(unflagged)

        # if ILLCOND may have been raised due to low equation counts
        if self._update_gain_flags("ILLCOND", flags_arr):
            unflagged = flags_arr == 0

        # gains flagged for any reason
        self.flagged = self.gflags != 0
        self.n_flagged = self.flagged.sum()

        return unflagged

        # # get error estimate model
        # model = np.sqrt(self.interval_sum(modelsq,1))
        # self.model_error = model*self.gain_error*(2+self.gain_error)

    def _report_gain_flags(self, whats_flagged, why_flagged="on low SNR",
                           check_settings="your max-prior-error settings",
                           threshold=0):
        def issue_warning(msg, sort_index=70):
            msg = "{0:s} {1:s}: {2:s}. " \
                  "You may need to check {3:s} and/or solution intervals. " \
                  "New flags will be raised for this chunk of data".format(
                self.chunk_label, self.jones_label, msg, check_settings)
            self.raise_userwarning(logging.WARNING, msg, sort_index, verbosity=log.verbosity())

        if whats_flagged.all():
            issue_warning("all solutions flagged {}".format(why_flagged))

        else:
            # whats_flagged has shape dir, time, freq, ant
            # check for various slices that are all-flagged and issue warnings
            if whats_flagged.all(axis=(0, 1, 3)).any():
                issue_warning("some frequency intervals flagged {}".format(why_flagged))
            if whats_flagged.all(axis=(0, 2, 3)).any():
                issue_warning("some time intervals flagged {}".format(why_flagged))
            # bad stations
            bad_stations = np.argwhere(whats_flagged.all(axis=(0, 1, 2))).flatten()
            if len(bad_stations):
                issue_warning("(faulty?) stations {0:s} ({1:d}/{2:d}) flagged {3:s}".format(
                    ", ".join(map(str, bad_stations)),
                    len(bad_stations), self.n_ant, why_flagged))
            # percentage flagged per direction
            percflagged = whats_flagged.sum(axis=(1, 2, 3)) * 100.0 / whats_flagged[0].size
            bad_dirs = np.argwhere(percflagged > threshold)

            if len(bad_dirs):
                if log.verbosity() > 1:
                    def __extract_dirs(d):
                        if (type(d) is list or type(d) is np.ndarray) and len(d) == 1:
                            return str(d[0])
                        elif (type(d) is list or type(d) is np.ndarray):
                            return ", ".join(d)
                        else:
                            return str(d)

                    def __extract_dirs_perc(d):
                        if (type(d) is list or type(d) is np.ndarray) and len(d) == 1:
                            return "{0:.3f}%".format(percflagged[d[0]])
                        elif (type(d) is list or type(d) is np.ndarray):
                            return ", ".join(["{0:.3f}%".format(percflagged[dd]) for dd in d])
                        else:
                            return "UNDETERMINED"

                    msg = "{} directions ({}) flagged {}".format(
                        len(bad_dirs),
                        ", ".join(["dir {0:s}: {1:s} gains affected".format(
                                  __extract_dirs(d), __extract_dirs_perc(d)) 
                            for d in bad_dirs]),
                        why_flagged)
                else:
                    msg = "{} directions flagged {}. Verbosity level 2 gives a full list of these directions".format(
                        len(bad_dirs), why_flagged)
                issue_warning(msg, 50)

    def _update_gain_flags(self, flagtype, flags_arr):
        """Helper function. Propagates flags given by mask onto data flags (if this is enabled),
        and updates counters. Returns number of gains flagged by this mask."""
        flagmask = getattr(FL, flagtype)
        flagged_gains = ((self.gflags & flagmask) != 0)

        self.num_flagged_on[flagtype] = num_flagged = flagged_gains.sum()

        if num_flagged and self.propagates_flags:
            bad_intervals = flagged_gains.any(axis=0) if self.propagates_anydir_flags else flagged_gains.all(axis=0)
            if bad_intervals.any():
                bad_slots = self.unpack_intervals(bad_intervals)
                bad_slots = bad_slots[:, :, :, np.newaxis]|bad_slots[:, :, np.newaxis, :]
                flags_arr[bad_slots] |= flagmask

        return num_flagged


    def update_equation_counts(self, unflagged):
        """Sets up equation counters based on flagging information. Overrides base version to compute
        additional stuff"""
        MasterMachine.update_equation_counts(self, unflagged)

        # The above computes:
        # Equations per antenna (summed over all time/freq slots)
        #    self.eqs_per_antenna = np.sum(unflagged, axis=(0, 1, 2)) * self.eqs_per_record
        # Total equations per time/freq slot (summed over all baselines)
        #    self.eqs_per_tf_slot = np.sum(unflagged, axis=(-1, -2)) * self.eqs_per_record

        # Equations per antenna and time/freq slot (summed over baselines)
        self.eqs_per_tf_slot_ant  = unflagged.sum(axis=-1) * self.eqs_per_record
        # Equations per antenna and solution interval
        self.eqs_per_interval_ant = self.interval_sum(self.eqs_per_tf_slot_ant)

        # First time 'round, set up missing intervals
        if self.missing_intervals_ant is None:
            # missing interval/antennas -- we can't solve for these, flag them as MISSING
            self.missing_intervals_ant = self.eqs_per_interval_ant == 0
            if self.solvable:
                self.gflags[:, self._interval_to_gainres(self.missing_intervals_ant)] = FL.MISSING

        # Mask of valid time/frequency intervals/antennas (i.e. ones for which we have enough equations)
        # This may be a subset of ~missing_intervals_ant, if some intervals have more than zero but insufficient equations

        self.valid_intervals_ant = self.eqs_per_interval_ant >= self.dof_per_antenna
        self.num_valid_intervals_ant = self.valid_intervals_ant.sum()

        # any interval/antennas that weren't missing (0 equations) to begin with, but don't have equations
        # now are marked ILLCOND
        illcond = ~(self.valid_intervals_ant | (self.eqs_per_interval_ant==0))
        self.gflags[:, self._interval_to_gainres(illcond)] = FL.ILLCOND

        # Now work out the chi-sq normalization factors

        # Total equations per interval
        eqs_per_interval = self.interval_sum(self.eqs_per_tf_slot)
        ndir = self.n_dir - len(self.fix_directions) if self.dd_term else 1
        # Number of unknowns per interval
        unknowns_per_interval = self.dof_per_antenna*self.n_ant*ndir

        valid_intervals = eqs_per_interval > unknowns_per_interval

        if valid_intervals.any():
            # Adjust chi-sq normalisation based on DoF count: MasterMachine computes chi-sq normalization
            # as 1/N_eq, we want to compute it as the reduced chi-square statistic, 1/(N_eq-N_dof)
            # This results in a per-interval correction factor

            with np.errstate(invalid='ignore', divide='ignore'):
                corrfact = eqs_per_interval.astype(float)/(eqs_per_interval - unknowns_per_interval)
            corrfact[~valid_intervals] = 0

            self._chisq_tf_norm_factor = self._chisq_tf_norm_factor_0 * self.unpack_intervals(corrfact)
            self._chisq_norm_factor = self._chisq_norm_factor * corrfact.sum() / valid_intervals.sum()


    def next_iteration(self):
        np.copyto(self.old_gains, self.gains)
        return MasterMachine.next_iteration(self)

    def flag_solutions(self, flags_arr, final=0):
        """ Flags gain solutions based on certain criteria, e.g. out-of-bounds, null, etc. """

        # Anything previously flagged for another reason will not be reflagged.

        flagged = self.flagged
        nfl0 = self.n_flagged

        # flag on OOB and such
        gain_mags = np.abs(self.gains)

        # Check for inf/nan solutions. One bad correlation will trigger flagging for all correlations.

        boom = (~np.isfinite(self.gains)).any(axis=(-1,-2))
        self.gflags[boom&~flagged] |= FL.BOOM
        flagged |= boom
        gain_mags[boom] = 0
        self._update_gain_flags("BOOM", flags_arr)

        # Check for gain solutions for which diagonal terms have gone to 0.

        gnull = (self.gains[..., 0, 0] == 0) | (self.gains[..., 1, 1] == 0)
        self.gflags[gnull&~flagged] |= FL.GNULL
        flagged |= gnull
        self._update_gain_flags("GNULL", flags_arr)

        # Check for gain solutions which are out of bounds (based on clip thresholds).
        if self.clip_after <= self.iters and (self.clip_upper or self.clip_lower):
            goob = np.zeros(gain_mags.shape, bool)
            if self.clip_upper:
                goob = gain_mags.max(axis=(-1, -2)) > self.clip_upper
            if self.clip_lower:
                goob |= (gain_mags[...,0,0]<self.clip_lower) | (gain_mags[...,1,1,]<self.clip_lower)
            self.gflags[goob&~flagged] |= FL.GOOB
            flagged |= goob
            self._update_gain_flags("GOOB", flags_arr)

        # in final (post-solution) flagging, check the posterior error estimate
        if final>0:
            if self.posterior_gain_error is not None and self.max_post_error:
                # reset to 0 for fixed directions
                if self.dd_term:
                    self.posterior_gain_error[self.fix_directions, ...] = 0
                # flag gains on max error
                # the last axis is correlation -- we flag if any correlation is flagged
                bad_gain_intervals = (self.posterior_gain_error > self.max_post_error).any(axis=(-1,-2))  # dir,time,freq,ant

                # mask high-variance gains that are not already otherwise flagged
                pge_flags = mask = self._interval_to_gainres(bad_gain_intervals, 1)&~flagged
                # raise FL.GVAR flag on these gains (and clear on all others!)
                self.gflags &= ~FL.GVAR
                self.gflags[mask] |= FL.GVAR
                flagged[mask] = True

                self._n_flagged_on_max_posterior_error = mask.sum(axis=(1, 2, 3)) if mask.any() else None

                if pge_flags.all(axis=0).all():
                    msg = "'{0:s}' {1:s} All directions flagged by posterior gain variance. This probably indicates significant RFI / outliers"\
                          "You need to check your max-post-error setting and data for selected intervals. "\
                          "New flags will be raised for this chunk of data".format(
                                self.jones_label, self.chunk_label)
                    self.raise_userwarning(logging.CRITICAL, msg, 70, verbosity=log.verbosity(), color="red")

                else:
                    dir_snr = {}
                    for d in range(pge_flags.shape[0]):
                        percflagged = np.sum(pge_flags[d]) * 100.0 / pge_flags[d].size
                        if percflagged > self.high_gain_var_warn and d not in self.fix_directions: dir_snr[d] = percflagged
                    if len(dir_snr) > 0:
                        if log.verbosity() > 2:
                            msg = "Signficiant gain variance in one or more directions of gain '{0:s}' chunk '{1:s}':".format(
                                    self.jones_label, self.chunk_label) +\
                                  "\n{0:s}\n".format("\n".join(["\t direction {0:s}: {1:.3f}% gains affected".format(
                                                        str(d), dir_snr[d]) for d in sorted(dir_snr)])) +\
                                  "Check your setting for max-post-error and your data for this interval. "\
                                  "New flags will be raised for this chunk. "
                        else:
                            msg = "'{0:s}' {1:s} Significant gain variance in directions {2:s}. "\
                                  "Check your data for this interval or raise max-post-error! "\
                                  "New flags will be raised for this chunk.".format(
                                self.jones_label, self.chunk_label, ", ".join(map(str, sorted(dir_snr))))
                        self.raise_userwarning(logging.CRITICAL, msg, 50, verbosity=log.verbosity(), color="red")

                    if pge_flags.all(axis=0).all(axis=0).all(axis=-1).any():
                        msg = "'{0:s}' {1:s} All time of one or more frequency intervals flagged due to gain variance. "\
                              "You need to check your max-post-error and data for this interval. "\
                              "New flags will be raised for this chunk of data".format(
                                    self.jones_label, self.chunk_label)
                        self.raise_userwarning(logging.WARNING, msg, 70, verbosity=log.verbosity())

                    if pge_flags.all(axis=0).all(axis=1).all(axis=-1).any():
                        msg = "'{0:s}' {1:s} All channels of one or more time intervals flagged due to gain variance. "\
                              "You need to check your max-post-error and data for this interval. "\
                              "New flags will be raised for this chunk of data".format(
                                    self.jones_label, self.chunk_label)
                        self.raise_userwarning(logging.WARNING, msg, 70, verbosity=log.verbosity())
                    stationflags = np.argwhere(pge_flags.all(axis=0).all(axis=0).all(axis=0)).flatten()
                    if stationflags.size > 0:
                        msg = "'{0:s}' {1:s} Stations {2:s} ({3:d}/{4:d}) fully flagged due to gain variation. "\
                              "These stations may be faulty or your variation requirements (max-post-error) are not met. "\
                              "New flags will be raised for this chunk of data".format(
                                    self.jones_label, self.chunk_label, ", ".join(map(str, stationflags)),
                                    np.sum(pge_flags.all(axis=0).all(axis=0).all(axis=0)), pge_flags.shape[3])
                        self.raise_userwarning(logging.WARNING, msg, 70, verbosity=log.verbosity())


                # if bad_gain_intervals.any():
                #     # (n_dir,) array showing how many were flagged per direction
                #     self._n_flagged_on_max_error = bad_gain_intervals.sum(axis=(1, 2, 3))
                #     # raised corresponding gain flags
                #     mask = self._interval_to_gainres(bad_gain_intervals, 1)
                #     self.gflags[mask] |= FL.GVAR
                #     flagged[mask] = True
                # else:
                #     self._n_flagged_on_max_posterior_error = None

        # Count the gain flags, excluding those set a priori due to missing data.

        self.flagged = self.gflags != 0
        self.n_flagged = flagged.sum()

        # keep flagged gains at their previous values
        self.gains[self.flagged] = self.old_gains[self.flagged]

        # propagate flags out to data if something new has been flagged, or if final<0 (which means
        # we're applying solutions, so better propagate everything)
        if (self.n_flagged > nfl0 or final<0) and self.propagates_flags:
            # convert gain flags to full time/freq resolution, and add directions together
            nodir_flags = self._gainres_to_fullres(np.bitwise_or.reduce(self.gflags, axis=0))

            # We remove the FL.MISSING bit when propagating as this bit is pre-set for data flagged
            # as PRIOR|MISSING. This prevents every PRIOR but not MISSING flag from becoming MISSING.

            flags_arr |= nodir_flags[:,:,:,np.newaxis]&~FL.MISSING
            flags_arr |= nodir_flags[:,:,np.newaxis,:]&~FL.MISSING

        if self.n_flagged != nfl0:
            self.update_equation_counts(flags_arr == 0)
            return True

        return False


    def num_gain_flags(self, mask=None, final=False):
        return int((self.gflags&(mask or ~FL.MISSING) != 0).sum()), self.gflags.size

    @property
    def dof_per_antenna(self):
        """This property returns the number of real degrees of freedom per antenna, per solution interval"""
        if "diag" in self.update_type:
            dofs = 2
        elif "scalar" in self.update_type:
            dofs = 1
        else:
            dofs = 4
        if "amp" not in self.update_type and "phase" not in self.update_type:
            dofs *= 2

        return dofs

    @property
    def has_valid_solutions(self):
        return (self.gflags==0).any()

    @property
    def num_solutions(self):
        return self.n_tf_ints

    @property
    def num_converged_solutions(self):
        return self.n_cnvgd

    def check_convergence(self, min_delta_g):
        """
        Updates the convergence parameters of the current time-frequency chunk. 

        Args:
            min_delta_g (float):
                Threshold for the minimum change in the gains - convergence criterion.
        """
        
        diff_g = np.square(np.abs(self.old_gains - self.gains))
        diff_g[self.flagged] = 0
        diff_g = diff_g.sum(axis=(-1,-2,-3))
        
        norm_g = np.square(np.abs(self.gains))
        norm_g[self.flagged] = 1
        norm_g = norm_g.sum(axis=(-1,-2,-3))

        norm_diff_g = diff_g/norm_g

        self.max_update = np.sqrt(np.max(diff_g))
        self.n_cnvgd = (norm_diff_g <= min_delta_g**2).sum()
        self._frac_cnvgd = self.n_cnvgd / float(norm_diff_g.size)

    def restrict_solution(self, gains):
        """
        Restricts the solutions by, for example, selecting a reference antenna or taking only the 
        amplitude. 
        """
        # raise flag so updates of G^H and G^-1 are computed
        self._gh_update = self._ghinv_update = True

        if "diag" in self.update_type or "scalar" in self.update_type:
            gains[...,(0,1),(1,0)] = 0

        if "scalar" in self.update_type:
            gains[...,(0,1),(0,1)] = gains[...,(0,1),(0,1)].mean(axis=-1)[...,np.newaxis]

        if "phase" in self.update_type:
            gabs = abs(gains)
            gnull = gabs==0
            with np.errstate(invalid='ignore'):
                gains /= gabs
            gains[gnull] = 0

        if "amp" in self.update_type:
            np.abs(gains, out=gains.real)
            gains.imag[:] = 0
        
        ## explicitly roll back invalid gains to previously known good values
        #self.gains[self.gflags != 0] = self.old_gains[self.gflags != 0]
        
        # explicitly roll back gains to previously known good values for fixed directions
        for idir in self.fix_directions:
            gains[idir, ...] = self.old_gains[idir, ...]
            # This might not be initialized in the load_from case.
            if self.posterior_gain_error is not None:
                self.posterior_gain_error[idir, ...] = 0

    @staticmethod
    def copy_or_identity(array, time_ind=0, out=None):
        """Helper conversion method. Returns array itself, or copies it to out"""
        if out is None:
            return array
        np.copyto(out, array)
        return out

    def unpack_intervals(self, arr, tdim_ind=0, out=None):
        """
        Helper conversion method. Unpacks an array that has time/freq axes in terms of intervals into a shape
        that has time/freq axes at full resolution.

        Args:
            arr (np.ndarray):
                Array with adjacent time and frequency axes.
            tdim_ind (int, optional):
                Position of time axis in array axes. 

        Returns:
            np.ndarray:
                Array at full resolution.
        """
        if out is not None:
            out[:] = arr[tuple([slice(None)] * tdim_ind + [self.t_mapping, self.f_mapping])]
            return out
        else:
            return arr[tuple([slice(None)] * tdim_ind + [self.t_mapping, self.f_mapping])]

    def interval_sum(self, arr, tdim_ind=0, out=None):
        """
        Helper conversion method. Sums an array with full resolution time/freq axes into time/freq intervals.

        Args:
            arr (np.ndarray):
                Array with adjacent time and frequency axes.
            tdim_ind (int, optional):
                Position of time axis in array axes. 

        Returns:
            np.ndarray:
                Array with interval resolution.
        """

        return np.add.reduceat(np.add.reduceat(arr, self.t_bins, tdim_ind), self.f_bins, tdim_ind+1, out=out)

    def interval_and(self, arr, tdim_ind=0, out=None):
        """
        Helper method. Logical-ands an array with full resolution time/freq axes into time/freq 
        intervals.

        Args:
            arr (np.ndarray):
                Array with adjacent time and frequency axes.
            tdim_ind (int, optional):
                Position of time axis in array axes. 

        Returns:
            np.ndarray:
                Array with interval resolution.
        """

        return np.logical_and.reduceat(np.logical_and.reduceat(arr, self.t_bins, tdim_ind), 
                                                                    self.f_bins, tdim_ind+1, out=out)
    @property
    def converged_fraction(self):
        return self._frac_cnvgd

    @property
    def has_converged(self):
        """ Returns convergence status. """
        return self.converged_fraction >= self.min_quorum or self.iters >= self.maxiter

    @has_converged.setter
    def has_converged(self, value):
        if not value:
            self._frac_cnvgd = self.n_cnvgd = 0

    @property
    def conditioning_status_string(self):
        """Returns conditioning status string"""
        if self.solvable:
            mineqs = self.eqs_per_interval_ant[self.valid_intervals_ant].min() if self.num_valid_intervals_ant else 0
            maxeqs = self.eqs_per_interval_ant.max()
            anteqs = (self.eqs_per_antenna!=0).sum()
            string = "{}: {}/{} ints".format(self.jones_label, self.num_valid_intervals_ant, self.n_tf_ints*self.n_ant)
            if self.num_valid_intervals_ant:
                string += " ({}-{} EPA)".format(mineqs, maxeqs)
                if self.dd_term:
                    string += " {} dirs".format(self.n_dir)
                string += " {}/{} ants, MGE {}".format(anteqs, self.n_ant,
                    " ".join(["{:.3}".format(self.prior_gain_error[idir, :].max()) for idir in range(self.n_dir)]))
                flagged = ["{}:{}".format(flagtype, count) for flagtype, count in self.num_flagged_on.items() if count]
                if flagged:
                    string += " (flagged {})".format(" ".join(flagged))

            return string
        else:
            return "{}: n/s".format(self.jones_label)


    @property
    def flagging_stats_string(self):
        """Returns a string describing per-flagset statistics"""
        fstats = []

        for flag, mask in FL.categories().items():
            n_flag = ((self.gflags & mask) != 0).sum()
            if n_flag:
                fstats.append("{}:{}({:.2%})".format(flag, n_flag, n_flag/float(self.gflags.size)))

        return " ".join(fstats)



    @property
    def current_convergence_status_string(self):
        """
        This property must return a status string for the gain machine, e.g.
            "G: 20 iters, conv 60.02%, g/fl 15.00%"
        """
        if self.solvable:
            string = "{}: {} iters, conv {:.2%}".format(self.jones_label, self.iters, self.converged_fraction)
            nfl, ntot = self.num_gain_flags()
            if nfl:
                string += ", g/fl {:.2%} [{}]".format(nfl/float(ntot), self.flagging_stats_string)
            if self.missing_gain_fraction:
                string += ", d/fl {:.2%}".format(self.missing_gain_fraction)
            string += ", max update {:.4}".format(self.max_update)
            if self.posterior_gain_error is not None:
                string += ", PGE " + " ".join(["{:.3}".format(self.posterior_gain_error[idir, :].max())
                                               for idir in range(self.n_dir)])
            return string
        else:
            return "{}: n/s{}".format(self.jones_label, ", loaded" if self._gains_loaded else "")

    @property
    def final_convergence_status_string(self):
        """
        This property must return a status string for the gain machine, e.g.
            "G: 20 iters, conv 60.02%, g/fl 15.00%"
        """
        if self.solvable:
            string = "{}: {} iters, conv {:.2%}".format(self.jones_label, self.iters, self.converged_fraction)
            nfl, ntot = self.num_gain_flags(final=True)
            if nfl:
                string += ", g/fl {:.2%} [{}]".format(nfl/float(ntot), self.flagging_stats_string)
            if self.missing_gain_fraction:
                string += ", d/fl {:.2%}".format(self.missing_gain_fraction)
            if self.posterior_gain_error is not None:
                string += ", PGE " + " ".join(["{:.3}".format(self.posterior_gain_error[idir, :].max())
                                               for idir in range(self.n_dir)])
            flagged = ["{}:{}".format(flagtype, count) for flagtype, count in self.num_flagged_on.items() if count]
            if flagged:
                string += " (flagged {})".format(" ".join(flagged))
            return string
        else:
            return "{}: n/s{}".format(self.jones_label, ", loaded" if self._gains_loaded else "")

    @property
    def has_stalled(self):
        """ Returns stalled status. """

        return self._has_stalled

    @has_stalled.setter
    def has_stalled(self, value):
        """ Sets stalled status. """

        self._has_stalled = value



#!/usr/bin/env python
from __future__ import print_function
from cubical.tools import logger
from collections import OrderedDict
import matplotlib.patches as mpatches
from pylab import *
import re, fnmatch
import warnings

log = logger.getLogger("gain_plots")


class PlotLimits(object):
    def __init__(self):
        self.max_reim = None
        self.max_reim_1 = None
        self.max_phase = None
        self.min_ampl = None
        self.max_ampl = None
        self.min_ampl_1 = None
        self.max_ampl_1 = None
        self.min_freq = None
        self.max_freq = None
        self.min_time = None
        self.max_time = None

class PlotOptions(PlotLimits):
    """This holds a set of common plotting options"""
    def __init__(self):
        self.dpi = 150
        self.font_size = 10
        self.ncol = 4
        self.nrow = 8
        self.width = 4
        self.height = 3

    @staticmethod
    def populate_argparse(parser):
        parser.add_argument("--ant", type=str, default='*', metavar="ANT",
                                help="Antenna subset. You may specify multiple comma-separated names or globs")
        parser.add_argument("--dir", type=int, default=0, metavar="DIR",
                                help="Direction number, for direction-dependent tables")
        parser.add_argument("--nrow", type=int, default=7, help="Plot rows")
        parser.add_argument("--ncol", type=int, default=4, help="Plot columns")
        parser.add_argument("--width", type=float, metavar="INCHES", default=4, help="Plot width")
        parser.add_argument("--height", type=float, metavar="INCHES", default=3, help="Plot height")
        parser.add_argument("--dpi", type=int, default=150, help="Plot DPI")
        parser.add_argument("--font-size", type=int, default=6, metavar="POINTS", help="Font size")

        parser.add_argument("--margin-left", type=float, default=0.05, metavar="FRAC", help="Left margin")
        parser.add_argument("--margin-right", type=float, default=0.05, metavar="FRAC", help="Right margin")
        parser.add_argument("--margin-bottom", type=float, default=0.05, metavar="FRAC", help="Bottom margin")
        parser.add_argument("--margin-top", type=float, default=0.05, metavar="FRAC", help="Top margin")

        parser.add_argument("--min-freq", type=float, metavar="MHz", default=None, help="Start frequency")
        parser.add_argument("--max-freq", type=float, metavar="MHz", default=None, help="End frequency")
        parser.add_argument("--min-time", type=float, metavar="s", default=None, help="Start time")
        parser.add_argument("--max-time", type=float, metavar="s", default=None, help="End time")

        parser.add_argument("--max-reim", type=float, metavar="VALUE", default=None, help="Sets re/im axis limits")
        parser.add_argument("--max-reim-1", type=float, metavar="VALUE", default=None, help="Sets re/im axis limits for off-diagonals")
        parser.add_argument("--max-phase", type=float, metavar="DEG", default=None, help="Sets phase axis limits")
        parser.add_argument("--min-ampl", type=float, metavar="AMPL", default=None, help="Sets ampl axis limits")
        parser.add_argument("--max-ampl", type=float, metavar="AMPL", default=None, help="Sets ampl axis limits")
        parser.add_argument("--min-ampl-1", type=float, metavar="AMPL", default=None, help="Sets ampl axis limits for off-diagonals")
        parser.add_argument("--max-ampl-1", type=float, metavar="AMPL", default=None, help="Sets ampl axis limits for off-diagonals")

options = PlotOptions()

def get_plot_limits(options, sols, time0=0):
    # get global plot limits
    auto_limits = PlotLimits()
    auto_limits.min_time = 0
    auto_limits.max_time = (max([max(time) for time,_,_,_,_,_ in sols.values() if len(time)] or [0]) - time0)/3600.
    auto_limits.min_freq = min([min(freq) for _,freq,_,_,_,_ in sols.values() if len(freq)] or [0])*1e-6
    auto_limits.max_freq = max([max(freq) for _,freq,_,_,_,_ in sols.values() if len(freq)] or [0])*1e-6
    if auto_limits.max_time == auto_limits.min_time:
        auto_limits.max_time += 1
    if auto_limits.max_freq == auto_limits.min_freq:
        auto_limits.max_freq += 1

    with warnings.catch_warnings():  # abs tends to raise complex warnings on masked arrays.
        warnings.simplefilter("ignore", ComplexWarning)
        max_ampls_diag = { ant: max(abs(g00).max(), abs(g11).max()) for ant, (_, _, g00, g01, g10, g11) in sols.items() }
        log.print("max diag amplitude per antenna:", ", ".join([ "{}: {}".format(ant, val) for ant, val in sorted(max_ampls_diag.items())]))
        auto_limits.max_ampl = max(max_ampls_diag.values())

        max_ampls_offdiag = { ant: max(abs(g01).max(), abs(g10).max()) for ant, (_, _, g00, g01, g10, g11) in sols.items() }
        log.print("max off-diag amplitude per antenna:", ", ".join([ "{}: {}".format(ant, val) for ant, val in sorted(max_ampls_offdiag.items())]))
        auto_limits.max_ampl_1 = max(max_ampls_offdiag.values())

        auto_limits.min_ampl = min([ min(abs(g00).min(), abs(g11).min()) for _, _, g00, g01, g10, g11 in sols.values()])
        auto_limits.min_ampl_1 = min([ min(abs(g01).min(), abs(g10).min()) for _, _, g00, g01, g10, g11 in sols.values()])
        auto_limits.max_phase = max([ max([abs(np.angle(g)).max() for g in ggs[2:]]) for ggs in sols.values()])*180/np.pi
        auto_limits.max_reim = max([ max(abs(g00.real).max(), abs(g00.imag).max(), abs(g11.real).max(), abs(g11.imag).max())
                                    for _, _, g00, g01, g10, g11 in sols.values()])
        auto_limits.max_reim_1 = max([ max(abs(g01.real).max(), abs(g01.imag).max(), abs(g10.real).max(), abs(g10.imag).max())
                                    for _, _, g00, g01, g10, g11 in sols.values()])
    # override with options
    for attr in dir(auto_limits):
        if attr.startswith("min") or attr.startswith("max"):
            val = getattr(options, attr)
            if val is not None:
                setattr(auto_limits, attr, val)

    return auto_limits

def get_plot_counts(nsols, plots_per_sol):
    num_plots = nsols*plots_per_sol
    rows_needed = int(math.ceil(num_plots / float(options.ncol)))
    log.info("{}x{} plots in {} columns need {} rows".format(nsols, plots_per_sol,options.ncol, rows_needed))
    if rows_needed > options.nrow:
        max_sols = options.nrow * options.ncol // plots_per_sol
        log.warn("Not enough rows specified, displaying only first {} solutions".format(max_sols))
    else:
        max_sols = nsols
    return max_sols, max_sols*plots_per_sol, rows_needed



def plot_bandpass(sols, plot_diag='ap', plot_offdiag='', gaintype=("Bandpass", "Offdiag bandpass"), figtitle=None):
    """
    Makes leakage plots given by sols dict
    """
    fig = figure(figsize=(options.ncol*options.width, options.nrow*options.height), dpi=options.dpi)
    tight_layout()
    subplots_adjust(left=options.margin_left, bottom=options.margin_bottom,
                    right=1-options.margin_right, top=1-options.margin_top)
    if figtitle:
        fig.suptitle(figtitle, fontsize=options.font_size*2)
    fontdict = dict(fontsize=options.font_size)
    fontdict_title = dict(weight='bold', fontsize=options.font_size * 1.2)

    nplot = 0

    lim = get_plot_limits(options, sols)
    x1 = list(sols.values())[0][2]
    if x1.shape[0] > 100:
        log.warn("making bandpass plot for {} time slices, this better be intentional!".format(x1.shape[0]))

    have_phase = 'ap' in set([plot_diag, plot_offdiag])
    max_sols, max_plots, rows_needed = get_plot_counts(len(sols), int(plot_diag in ('ap', 'ri')) + int(plot_offdiag in ('ap', 'ri')))

    def _prep_plot(nplot, ant, num_gaintype, plot_type):
        nplot += 1
        ax = subplot(options.nrow, options.ncol, nplot)
        last_row = (max_plots - nplot) < options.ncol

        title("{} antenna {}".format(gaintype[num_gaintype], ant), fontdict=fontdict_title)

        ax.set_xlim(lim.min_freq, lim.max_freq)
        ax.tick_params("both", direction="in", labelsize=options.font_size)
        ax2 = ax.twinx()
        ax2.set_ylim(-lim.max_phase, lim.max_phase)
        #ax2.set_xticks([])
        ax2.tick_params("y", direction="in", labelsize=options.font_size)
        if last_row:
            ax.set_xlabel("Frequency (MHz)", fontdict=fontdict)
        else:
            ax.set_xticklabels([])
        # label phase axis if needed
        if have_phase and nplot % options.ncol == 0:
            ax2.set_ylabel("Phase (deg)", fontdict=fontdict)
        else:
            ax2.set_yticklabels([])
            if plot_type != 'ap':
                ax2.set_yticks([])
        for axis in ax.xaxis, ax.yaxis:
            for tick in axis.get_major_ticks():
                tick.label.set_fontsize(options.font_size)
        ax.get_yaxis().get_major_formatter().set_useOffset(False)
        return nplot, ax, ax2

    def _make_reim_plot(ax, freq, x1, x2, corrs, legend):
        re1, im1, re2, im2 = "Re "+corrs[0], "Im "+corrs[0], "Re "+corrs[1], "Im "+corrs[1]
        for ts in range(x1.shape[0]):
            ax.plot(freq, x1[ts].real, '+r', ms=2, label=re1 if not ts else None)
            ax.plot(freq, x1[ts].imag, '+b', ms=2, label=im1 if not ts else None)
        for ts in range(x2.shape[0]):
            ax.plot(freq, x2[ts].real, '+c', ms=2, label=re1 if not ts else None)
            ax.plot(freq, x2[ts].imag, '+y', ms=2, label=im2 if not ts else None)
        if legend:
            ax.legend(handles=[mpatches.Patch(color=col, label=label) for col, label in
                                   (("r", re1), ("b", im1),
                                    ("c", re2), ("y", im2))
                               ], loc="upper center", ncol=2, fontsize=options.font_size)

    def _make_ap_plot(ax, ax2, freq, x1, x2, corrs, legend):
        amp1, ph1, amp2, ph2 = "|"+corrs[0]+"|", corrs[0]+ " phase", "|"+corrs[1]+"|", corrs[1]+" phase"
        for ts in range(x1.shape[0]):
            ax2.plot(freq, np.angle(x1[ts]) * 180 / math.pi, '.c', ms=0.5,
                     label=ph1 if not ts else None)
            ax2.plot(freq, np.angle(x2[ts]) * 180 / math.pi, '.y', ms=0.5,
                     label=ph2 if not ts else None)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", ComplexWarning)
                ax.plot(freq, abs(x1[ts]), '+r', ms=2, label=amp1 if not ts else None)
                ax.plot(freq, abs(x2[ts]), '+b', ms=2, label=amp2 if not ts else None)
        if legend:
            ax.legend(handles=[mpatches.Patch(color=col, label=label) for col, label in
                                   (("r", amp1), ("b", amp2),
                                    ("c", ph1),  ("y", ph2))
                               ], loc="upper center", ncol=2, fontsize=options.font_size)

    for iant, (ant, (time, freq, g00, g01, g10, g11)) in enumerate(sols.items()):
        if iant >= max_sols:
            break
        #    print "shape is {}, grid span is {} {}".format(d00.shape, time[[0,-1]], freq[[0,-1]])
        freq = freq * 1e-6

        if plot_diag in ('ap', 'ri'):
            nplot, ax, ax2 = _prep_plot(nplot, ant, 0, plot_diag)

            if plot_diag == 'ap':
                if nplot % options.ncol == 1:
                    ax.set_ylabel("Amplitude", fontdict=fontdict)
                ax.set_ylim(lim.min_ampl, lim.max_ampl)
                _make_ap_plot(ax, ax2, freq, g00, g11, ("RR", "LL"), legend=not iant)
                # label axis if needed
                #if all_ap and nplot % options.ncol > 1:
                #    ax.set_yticklabels([])
            else:
                ax.set_ylim(-lim.max_reim, lim.max_reim)
                _make_reim_plot(ax, freq, g00, g11, ("RR", "LL"), legend=not iant)
                # label axis if needed
                #if all_ri and nplot % options.ncol > 1:
                #    ax.set_yticklabels([])

        if plot_offdiag in ('ap', 'ri'):
            nplot, ax, ax2 = _prep_plot(nplot, ant, 1,  plot_offdiag)

            if plot_offdiag == 'ap':
                if nplot % options.ncol == 1:
                    ax.set_ylabel("Amplitude", fontdict=fontdict)
                ax.set_ylim(lim.min_ampl_1, lim.max_ampl_1)
                _make_ap_plot(ax, ax2, freq, g01, g10, ("RL", "LR"), legend=not iant)
            else:
                ax.set_ylim(-lim.max_reim_1, lim.max_reim_1)
                _make_reim_plot(ax, freq, g01, g10, ("RL", "LR"), legend=not iant)
                # label axis if needed
                #if all_ri and nplot % options.ncol > 1:
                #    ax.set_yticklabels([])


    log.print("{} plots generated ({} rows and {} columns)".format(nplot, options.nrow, options.ncol))
    return fig

def plot_gain(sols, plot_diag='ap', plot_offdiag='', gaintype=("Gain", "Offdiag gain"), figtitle=None):
    """
    Makes leakage plots given by sols dict
    """
    fig = figure(figsize=(options.ncol*options.width, options.nrow*options.height), dpi=options.dpi)
    tight_layout()
    subplots_adjust(left=options.margin_left, bottom=options.margin_bottom,
                    right=1-options.margin_right, top=1-options.margin_top)
    if figtitle:
        fig.suptitle(figtitle, fontsize=options.font_size*2)
    fontdict = dict(fontsize=options.font_size)
    fontdict_title = dict(weight='bold', fontsize=options.font_size * 1.2)

    nplot = 0

    time0 = min([time.min() for time,_,_,_,_,_ in sols.values()])

    lim = get_plot_limits(options, sols, time0=time0)

    x1 = list(sols.values())[0][2]
    if x1.shape[1] > 100:
        log.warn("making gain plot for {} frequency slices, this better be intentional!".format(x1.shape[1]))

    have_phase = 'ap' in set([plot_diag, plot_offdiag])
    max_sols, max_plots, rows_needed = get_plot_counts(len(sols), int(plot_diag in ('r', 'ap', 'ri')) +
                                                                  int(plot_offdiag in ('r', 'ap', 'ri')))

    def _prep_plot(nplot, ant, num_gaintype, plot_type):
        nplot += 1
        ax = subplot(options.nrow, options.ncol, nplot)
        last_row = (max_plots - nplot) < options.ncol

        title("{}: antenna {}".format(gaintype[num_gaintype], ant), fontdict=fontdict_title)

        ax.set_xlim(lim.min_time, lim.max_time)
        ax.tick_params("both", direction="in", labelsize=options.font_size)
        ax2 = ax.twinx()
        ax2.set_ylim(-lim.max_phase, lim.max_phase)
        #ax2.set_xticks([])
        ax2.tick_params("y", direction="in", labelsize=options.font_size)
        if last_row:
            ax.set_xlabel("Time (h since start)", fontdict=fontdict)
        else:
            ax.set_xticklabels([])
        # label phase axis if needed
        if have_phase and nplot % options.ncol == 0:
            ax2.set_ylabel("Phase (deg)", fontdict=fontdict)
        else:
            ax2.set_yticklabels([])
            if plot_type != 'ap':
                ax2.set_yticks([])
        for axis in ax.xaxis, ax.yaxis:
            for tick in axis.get_major_ticks():
                tick.label.set_fontsize(options.font_size)
        ax.get_yaxis().get_major_formatter().set_useOffset(False)
        return nplot, ax, ax2

    def _make_real_plot(ax, time, x1, x2, corrs, legend):
        for fs in range(x1.shape[1]):
            ax.plot(time, x1[:, fs], '+r', ms=2, label=corrs[0] if not fs else None)
        for fs in range(x2.shape[1]):
            ax.plot(time, x2[:, fs], '+b', ms=2, label=corrs[1] if not fs else None)
        if legend:
            ax.legend(handles=[mpatches.Patch(color=col, label=label) for col, label in
                                   (("r", corrs[0]), ("b", corrs[1]))
                               ], loc="upper center", ncol=2, fontsize=options.font_size)

    def _make_reim_plot(ax, time, x1, x2, corrs, legend):
        re1, im1, re2, im2 = "Re "+corrs[0], "Im "+corrs[0], "Re "+corrs[1], "Im "+corrs[1]
        for fs in range(x1.shape[1]):
            ax.plot(time, x1[:, fs].real, '+r', ms=2, label=re1 if not fs else None)
            ax.plot(time, x1[:, fs].imag, '+b', ms=2, label=im1 if not fs else None)
        for fs in range(x2.shape[1]):
            ax.plot(time, x2[:, fs].real, '+c', ms=2, label=re1 if not fs else None)
            ax.plot(time, x2[:, fs].imag, '+y', ms=2, label=im2 if not fs else None)
        if legend:
            ax.legend(handles=[mpatches.Patch(color=col, label=label) for col, label in
                                   (("r", re1), ("b", im1),
                                    ("c", re2), ("y", im2))
                               ], loc="upper center", ncol=2, fontsize=options.font_size)

    def _make_ap_plot(ax, AX2, time, x1, x2, corrs, legend):
        amp1, ph1, amp2, ph2 = "|"+corrs[0]+"|", corrs[0]+ " phase", "|"+corrs[1]+"|", corrs[1]+" phase"
        for fs in range(x1.shape[1]):
            ax2.plot(time, np.angle(x1[:, fs]) * 180 / math.pi, '.c', ms=0.5,
                     label=ph1 if not fs else None)
            ax2.plot(time, np.angle(x2[:, fs]) * 180 / math.pi, '.y', ms=0.5,
                     label=ph2 if not fs else None)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", ComplexWarning)
                ax.plot(time, abs(x1[:, fs]), '+r', ms=2, label=amp1 if not fs else None)
                ax.plot(time, abs(x2[:, fs]), '+b', ms=2, label=amp2 if not fs else None)
        if legend:
            ax.legend(handles=[mpatches.Patch(color=col, label=label) for col, label in
                                   (("r", amp1), ("b", amp2),
                                    ("c", ph1),  ("y", ph2))
                               ], loc="upper center", ncol=2, fontsize=options.font_size)

    for iant, (ant, (time, freq, g00, g01, g10, g11)) in enumerate(sols.items()):
        #    print "shape is {}, grid span is {} {}".format(d00.shape, time[[0,-1]], freq[[0,-1]])
        time = (time-time0)/3600.

        if nplot >= max_plots:
            break

        if plot_diag in ('ap', 'ri', 'r'):
            nplot, ax, ax2 = _prep_plot(nplot, ant, 0, plot_diag)

            if plot_diag == 'ap':
                if nplot % options.ncol == 1:
                    ax.set_ylabel("Amplitude", fontdict=fontdict)
                ax.set_ylim(lim.min_ampl, lim.max_ampl)
                _make_ap_plot(ax, ax2, time, g00, g11, ("RR", "LL"), legend=not iant)
            elif plot_diag == 'ri':
                ax.set_ylim(-lim.max_reim, lim.max_reim)
                _make_reim_plot(ax, time, g00, g11, ("RR", "LL"), legend=not iant)
            elif plot_diag == 'r':
                ax.set_ylim(-lim.max_reim, lim.max_reim)
                _make_real_plot(ax, time, g00, g11, ("RR", "LL"), legend=not iant)

        if nplot >= max_plots:
            break

        if plot_offdiag in ('ap', 'ri', 'r'):
            nplot, ax, ax2 = _prep_plot(nplot, ant, 1,  plot_offdiag)

            if plot_offdiag == 'ap':
                if nplot % options.ncol == 1:
                    ax.set_ylabel("Amplitude", fontdict=fontdict)
                ax.set_ylim(lim.min_ampl_1, lim.max_ampl_1)
                _make_ap_plot(ax, ax2, time, g01, g10, ("RL", "LR"), legend=not iant)
            elif plot_offdiag == 'ri':
                ax.set_ylim(-lim.max_reim_1, lim.max_reim_1)
                _make_reim_plot(ax, time, g01, g10, ("RL", "LR"), legend=not iant)
            elif plot_offdiag == 'r':
                ax.set_ylim(-lim.max_reim, lim.max_reim)
                _make_real_plot(ax, time, g00, g11, ("RL", "LR"), legend=not iant)

    log.print("{} plots generated ({} rows and {} columns)".format(nplot, options.nrow, options.ncol))
    return fig


def get_freq_slice(FS, all_freqs):
    if FS is None:
        if0 = if1 = None
        if options.min_freq is not None:
            wh = np.where(all_freqs >= options.min_freq*1e+6)[0]
            if0 = wh[0] if len(wh) else None
        if options.max_freq is not None:
            wh = np.where(all_freqs >= options.max_freq*1e+6)[0]
            if1 = wh[0] if len(wh) else None
        if if0 is not None or if1 is not None:
            FS = slice(if0 or 0, if1)
        else:
            FS = slice(None)
    return FS

def get_time_slice(TS, all_times):
    if TS is None:
        return slice(None)
    return TS


def _get_jones_element_index(G, ant, corr1=0, corr2=0, dir=0):
    index = dict(ant=ant)
    if 'dir' in G.axis_index:
        index.update(dir=dir)
    if 'corr1' in G.axis_index and 'corr2' in G.axis_index:
        index.update(corr1=corr1, corr2=corr2)
    elif corr1 != corr2:
        return None
    elif 'corr' in G.axis_index:
        index.update(corr=corr1)
    return index


def _get_jones_element_slice(G, ant, corr1=0, corr2=0, dir=0):
    index = _get_jones_element_index(G, ant, corr1, corr2, dir=dir)
    if index is None:
        return None, (None, None)
    return G.get_slice(**index)


def prepare_sols_dict(G, FS=None, TS=None):
    """
    Extract solutions slices into a dict
    """
    FS = get_freq_slice(FS, G.grid[G.ax.freq])
    TS = get_freq_slice(TS, G.grid[G.ax.time])
    have_offdiag = False
    is_complex = False

    sols = OrderedDict()

    patterns = options.ant.split(",") or ['*']

    for iant, ant in enumerate(G.grid[G.ax.ant]):
        if not any([fnmatch.fnmatch(ant, patt) for patt in patterns]):
            continue

        # this gets the "raw" solutions for a given slice (antenna, correlation, etc.), and also the grid they're defined on,
        # which could be a subset of the full grid given by the description
        g00, (time, freq) = _get_jones_element_slice(G, ant=iant, corr1=0, corr2=0, dir=options.dir)
        if g00 is None:
            continue
        g11, _ = _get_jones_element_slice(G, ant=iant, corr1=1, corr2=1, dir=options.dir)
        g01, _ = _get_jones_element_slice(G, ant=iant, corr1=0, corr2=1, dir=options.dir)
        g10, _ = _get_jones_element_slice(G, ant=iant, corr1=1, corr2=0, dir=options.dir)
        sols[ant] = time[TS], freq[FS], \
                        g00[TS, FS], \
                        (g01[TS, FS] if g01 is not None else np.array([0.])), \
                        (g10[TS, FS] if g10 is not None else np.array([0.])), \
                        g11[TS, FS]
        if g01 is not None or g10 is not None:
            have_offdiag = True
        if any([np.iscomplexobj(g) for g in sols[ant][2:]]):
            is_complex = True

    return sols, have_offdiag, is_complex

def plot_bandpass_cc(G, FS=None, TS=None,
                        plot_diag='ap', plot_offdiag='', figtitle=None, component='gain'):
    """Plots G solutions from a CubiCal database"""
    sols, have_offdiag, is_complex = prepare_sols_dict(G, FS, TS)
    if not have_offdiag:
        plot_offdiag = ''
    if not is_complex:
        plot_diag = plot_diag and 'r'
        plot_offdiag = plot_offdiag and 'r'
    if component == 'gain':
        gaintype = 'Bandpass', 'Offdiag bandpass'
    else:
        gaintype = component.capitalize(), 'Offdiag ' + component
    return plot_bandpass(sols, plot_diag=plot_diag, plot_offdiag=plot_offdiag, gaintype=gaintype, figtitle=figtitle)

def plot_gain_cc(G, FS=None, TS=None,
                        plot_diag='ap', plot_offdiag='', figtitle=None, component='gain'):
    """Plots G solutions from a CubiCal database"""
    sols, have_offdiag, is_complex = prepare_sols_dict(G, FS, TS)
    if not have_offdiag:
        plot_offdiag = ''
    if not is_complex:
        plot_diag = plot_diag and 'r'
        plot_offdiag = plot_offdiag and 'r'
    if component == 'gain':
        gaintype = 'Gain', 'Offdiag gain'
    else:
        gaintype = component.capitalize(), 'Offdiag ' + component
    return plot_gain(sols, plot_diag=plot_diag, plot_offdiag=plot_offdiag, gaintype=gaintype, figtitle=figtitle)


def read_cubical_gains(filename, label=None, component="gain"):
    """
    Reads CubiCal leakage solutions from a CubiCal DB
    """
    from cubical.database import pickled_db
    db = pickled_db.PickledDatabase()
    db._load(filename)
    log.print("{} contains solutions for {}".format(filename, " ".join(db.names())))
    # try to auto-pick a label
    if label is None:
        for name in db.names():
            match = re.match("(\\w+):{}".format(component), name)
            if match:
                label = match.group(1)
                log.print("  gain label is {}".format(label))
                break
        else:
            log.error("no gain entries found in database")
            return None
    gg = db['{}:{}'.format(label, component)]
    log.print("  loading {}:{} solutions".format(label, component))
    log.print("  axes are: {}; shape is {}".format(", ".join(gg.axis_labels), gg.shape))  # axis info
    log.print("  can be interpolated over axes {}".format(" ".join([gg.axis_labels[i] for i in gg.interpolation_axes])))
    time, freq = gg.grid[gg.ax.time], gg.grid[gg.ax.freq]  # grid info
    log.print("    time span {} h (start {:f}), freq span {} to {} MHz".format(
                    (time[-1]-time[0])/3600, time[0], freq[0]*1e-6, freq[-1]*1e-6))

    corr00 = {}
    for axis in 'corr', 'corr1', 'corr2':
        if axis in gg.axis_index:
            corr00[axis] = 0

    valid_ants = set([ant for iant, ant in enumerate(gg.grid[gg.ax.ant])
                      if gg.is_slice_valid(**_get_jones_element_index(gg, iant, 0, 0, dir=options.dir))])

    log.print("  valid antennas:", " ".join([ant for ant in gg.grid[gg.ax.ant] if ant in valid_ants]))
    log.print("  missing antennas:", " ".join([ant for ant in gg.grid[gg.ax.ant] if ant not in valid_ants] or ["none"]))
    return gg



# -*- coding: utf-8 -*-
# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details
from __future__ import print_function
from builtins import range
import numpy as np
import sys
import traceback
from collections import OrderedDict

from cubical.tools import shared_dict
from cubical.flagging import FL
from cubical import data_handler
from cubical.tools import dtype_checks as dtc


from cubical.tools import logger, ModColor
log = logger.getLogger("ms_tile")
try:
    from .TiggerSourceProvider import TiggerSourceProvider
except ImportError:
    TiggerSourceProvider = None
try:
    from cubical.degridder.DicoSourceProvider import DicoSourceProvider
except ImportError:
    DicoSourceProvider = None


## TERMINOLOGY:
## A "chunk" is data for one DDID, a range of timeslots (thus, a subset of the MS rows), and a
## slice of channels. Chunks are the basic parallelization unit. Solver deals with a chunk of data.
##
## A "row chunk" is data for one DDID, a range of timeslots, and *all* channels. One can imagine a
## row chunk as a "horizontal" vector of chunks across frequency.
##
## A "tile" is a collection of row chunks that are adjacent in time and/or DDID. One can imagine a
## tile as a vertical stack of row chunks

class RowChunk(object):
    """ Very basic helper class. Encapsulates a row chunk -- a set of rows representing a time interval, and a single
    DDID, split into a number of frequency chunks.

    A row chunk also knows about row rebinning.
    """

    def __init__(self, ddid, tchunk, timeslice, rows, rows0):
        """
        Initialises a RowChunk.

        Args:
            ddid (int):
                DDID index for the RowChunk.
            tchunk (int):
                Time index for the RowChunk.
            rows (np.ndarray):
                An (nrows_in_chunk) size array of rebinned row indices.
            rows0 (np.ndarray):
                An (nrows_in_chunk) size array of original row indices.
        """

        self.ddid, self.tchunk, self.timeslice, self.rows, self.rows0 = ddid, tchunk, timeslice, rows, rows0
        # MSTile.Subset to which this chunk belongs (a subset may contain several chunks)
        self.subset = None
        # subset_rows: row indices within a subset corresponding to this chunk
        self.rows_within_subset = None


class MSTile(object):
    """
    Helper class which encapsulates a tile. A tile is a sequence of row chunks that's read and
    written as a unit.
    """
    class Subset(object):
        """
        A Subset represents a part of the tile (one or more DDIDs) with the same channel structure.

        Subsets are read from the MS in one go, and predicted by Montblanc in one go.
        """
        def __init__(self, tile, label, datadict, ms_rows, ms_rows0):
            """
            Creates a subset object
            tile:       parent MSTile
            label:      a descriptive label for this subset
            datadict:   name of SharedDict correspondign to this subset
            ms_rows:    a slice (or list) of the rebinned MS rows comprising this subset. Row 0 is rebinned MS row 0.
            ms_rows0:   a list of the original MS rows comprising this subset. Row 0 is raw MS row 0.
            """
            self.tile = tile
            self.datadict = datadict
            self.rows0 = ms_rows0              # row numbers in original (sorted) MS
            self.label = label
            self.nants = tile.nants
            # 
            if type(ms_rows) is slice:
                log(1).print(f"{tile.label} subset {label} rows {ms_rows}")
            else:
                log(1).print(f"{tile.label} subset {label} contains {len(ms_rows)} rows (min {ms_rows[0]} max {ms_rows[-1]})")
            # subsets of rebinned rows
            self.ddid_col = self.tile.dh.ddid_col[ms_rows]
            self.time_col = self.tile.dh.time_col[ms_rows]
            self.antea = self.tile.dh.antea[ms_rows]
            self.anteb = self.tile.dh.anteb[ms_rows]
            self.times = self.tile.dh.times[ms_rows]
            self.nrows = len(self.times)
            # first DDID in subset
            self.first_ddid = self.ddid_col[0]
            self.nfreq = len(tile.dh.chanfreqs[self.first_ddid])
            self.ncorr = tile.dh.ncorr

            # row map for rebinning -- label of None means subset is full tile
            if label is None:
                self.rebin_row_map = tile.rebin_row_map
            else:
                self.rebin_row_map = tile.rebin_row_map[ms_rows0 - tile.first_row0]
                # since the rebinned map is now sparse in the output because we've got a subset of input rows,
                # uniquify it so it becomes contiguous in the output.
                # Note that '-' signs need to be preserved
                sign = np.sign(self.rebin_row_map)
                self.rebin_row_map, _, _ = data_handler.uniquify(abs(self.rebin_row_map))
                self.rebin_row_map *= sign

            # channel map for rebinning
            self.rebin_chan_map = tile.dh.rebin_chan_maps[self.first_ddid]
            if self.rebin_chan_map is None:
                self.rebin_chan_map = np.arange(0, len(tile.dh.chanfreqs[self.first_ddid]), dtype=np.int64)

            # filled in by self.load_montblanc_models below
            self._mb_measet_src = self._mb_cached_ms_src = None

            # bitflag column stored here
            self.bflagcol = self.bflagrow = self._flagcol = self_flagcol_sum = None

        # static member will be initialized with FitsBeamSourceProvider if needed
        _mb_arbeam_src = None

        def upsample(self, data):
            """Helper method. Upsamples an array back to full resolution"""
            if not self.tile.dh.do_freq_rebin and not self.tile.dh.do_time_rebin:
                return data
            shape = len(self.rebin_row_map), len(self.rebin_chan_map), data.shape[2]
            print("upsampling to {} rows and {} channels".format(shape[0], shape[1]), file=log(1))
            return data[self.rebin_row_map[:, np.newaxis],
                        self.rebin_chan_map[np.newaxis, :], None].reshape(shape)

        def load_ddfacet_models(self, uvwco, loaded_models, model_source, cluster, imod, idir, model_type):
            """
            Invoke DDFacet degridder to compute model visibilities
            for all directions.

            Postcondition of loaded_models is a new entry
            of model_source cluster object with all directions predicted.

            Args:
                uvwco: uvws for this tile
                loaded_models: dictionary of all loaded models including clusters
                model_source: cluster object
                cluster: cluster direction name to use as return value
                imod: index of this model provider (only used for printing)
                idir: index of global direction (only used for printing)
                model_type: Either 'cplx2x2' or 'cplxdiag' or 'cplxscalar' supported at the moment
            Returns:
                ndarray of shape nrow x nchan x ncorr for name specified in "cluster"
            """
            from cubical.degridder.DDFacetSim import DDFacetSim
            ddid_index, uniq_ddids, _ = data_handler.uniquify(self.ddid_col)
            self._freqs = np.array([self.tile.dh.chanfreqs[ddid] for ddid in uniq_ddids])
            self._chan_widths = np.array([self.tile.dh.chanwidth[ddid] for ddid in uniq_ddids])
            ddfsim = DDFacetSim()
            ndirs = model_source._nclus
            loaded_models[model_source] = {}
            ddfsim.set_model_provider(model_source)
            log.info("Predicting targeted directions... This may take some time")
            for idir, clus in enumerate(model_source._cluster_keys):
                ddfsim.set_direction(clus)
                model = ddfsim.simulate(self.tile.dh,
                                        self.tile,
                                        self,
                                        self.tile.dh._poltype,
                                        uvwco,
                                        self._freqs,
                                        model_type,
                                        self._chan_widths,
                                        self.time_col)
                loaded_models[model_source][clus] = model

            # finally return the direction requested in cluster
            model = loaded_models[model_source][cluster]
            print("  using {}{} for model {} direction {}".format(model_source,
                                "" if not cluster else ("()" if cluster == 'die' else "({})".format(cluster)),
                                imod, idir), file=log(1))

            return model

        def load_montblanc_models(self, uvwco, loaded_models, model_source, cluster, imod, idir):
            """
            Invoke Montblanc to compute model visibilities

            Args:
                uvwco:
                loaded_models:
                model_source:
                cluster:
                imod:
                idir:

            Returns:

            """

            from . import MBTiggerSim
            from montblanc.impl.rime.tensorflow.sources import CachedSourceProvider, FitsBeamSourceProvider
            assert dtc.assert_isint(self.times), self.times.dtype
            assert dtc.assert_isint(self.antea), self.antea.dtype
            assert dtc.assert_isint(self.nants)
            assert dtc.assert_isint(self.anteb), self.anteb.dtype
            assert dtc.assert_isfp(self.time_col), self.time_col.dtype
            assert dtc.assert_isint(self.ddid_col), self.ddid_col.dtype

            # setup montblanc machinery once per subset (may be called multiple times for different models)
            if not self._mb_measet_src:
                # massage data columns into Montblanc-friendly shapes
                # Given data, we need to make sure that it looks the way MB wants it to.
                # First step - check the number of rows.

                n_bl = (self.nants * (self.nants - 1)) // 2
                uniq_times = np.unique(self.times)
                ntime = len(uniq_times)
                uniq_time_col = np.unique(self.time_col)
                t_offset = uniq_times[0]

                # The row identifiers determine which rows in the SORTED/ALL ROWS are required for the data
                # that is present in the MS. Essentially, they allow for the selection of an array of a size
                # matching that of the observed data. First term determines the offset by ddid, the second
                # is the offset by time, and the last turns antea and anteb into a unique offset per
                # baseline.

                ddid_index, uniq_ddids, _ = data_handler.uniquify(self.ddid_col)

                self._freqs = np.array([self.tile.dh.chanfreqs[ddid] for ddid in uniq_ddids])


                assert dtc.assert_isint(n_bl)
                assert dtc.assert_isint(ddid_index)

                self._row_identifiers = ddid_index * n_bl * ntime + (self.times - self.times[0]) * n_bl + \
                                  (-0.5 * self.antea ** 2 + (self.nants - 1.5) * self.antea + self.anteb - 1).astype(
                                      np.int32)
                # make full list of row indices in Montblanc-compliant order (ddid-time-ant1-ant2)
                full_index = [(p, q, t, d) for d in range(len(uniq_ddids)) for t in uniq_times
                              for p in range(self.nants) for q in range(self.nants)
                              if p < q]

                self._expected_nrows = len(full_index)

                # and corresponding full set of indices
                full_row_set = set(full_index)

                print("  {} rows ({} expected for {} timeslots, {} baselines and {} DDIDs)".format(
                    self.nrows, self._expected_nrows, ntime, n_bl, len(uniq_ddids)), file=log(1))

                # make mapping from existing indices -> row numbers, omitting autocorrelations
                current_row_index = {(p, q, t, d): row for row, (p, q, t, d) in
                                     enumerate(zip(self.antea, self.anteb, self.times, ddid_index)) if p != q}

                # do we need to add fake rows for missing data?
                missing = full_row_set.difference(iter(current_row_index.keys()))
                nmiss = len(missing)

                if nmiss:
                    print("  {} rows will be padded in for Montblanc".format(nmiss), file=log(1))
                    # pad up columns
                    uvwco = np.concatenate((uvwco, [[0, 0, 0]] * nmiss))
                    antea = np.concatenate((self.antea,
                                            np.array([p for (p, q, t, d) in missing])))
                    anteb = np.concatenate((self.anteb,
                                            np.array([q for (p, q, t, d) in missing])))
                    time_col = np.concatenate((self.time_col,
                                               np.array([uniq_time_col[t - t_offset]
                                                         for (p, q, t, d) in missing])))
                    ddid_index = np.concatenate((ddid_index,
                                               np.array([d for (p, q, t, d) in missing])))
                    # extend row index
                    current_row_index.update({idx: (row + self.nrows) for row, idx in enumerate(missing)})
                else:
                    antea, anteb, time_col = self.antea, self.anteb, self.time_col

                # lookup each index in Montblanc order, convert it to a row number
                self._mb_sorted_ind = np.array([current_row_index[idx] for idx in full_index])

                self._mb_measet_src = MBTiggerSim.MSSourceProvider(self.tile, time_col, antea, anteb, ddid_index, uvwco,
                                                       self._freqs, self._mb_sorted_ind, len(self.time_col),
                                                       do_pa_rotation=self.tile.dh.pa_rotate_montblanc)

                if not self.tile.dh.pa_rotate_montblanc:
                    log.warn("Disabling LSM parallactic rotation as per user request")
                cache_data_sources = ["parallactic_angles"] if self.tile.dh.pa_rotate_montblanc else []
                self._mb_cached_ms_src = CachedSourceProvider(self._mb_measet_src,
                                                 cache_data_sources=cache_data_sources,
                                                 clear_start=False, clear_stop=False)
                if self.tile.dh.beam_pattern and self._mb_arbeam_src is None:
                    self._mb_arbeam_src = FitsBeamSourceProvider(self.tile.dh.beam_pattern,
                                                                 self.tile.dh.beam_l_axis,
                                                                 self.tile.dh.beam_m_axis)

            print("  computing visibilities for {}".format(model_source), file=log(0))
            # setup Montblanc computation for this LSM
            tigger_source = model_source
            cached_src = CachedSourceProvider(tigger_source, clear_start=True, clear_stop=True)
            srcs = [self._mb_cached_ms_src, cached_src]

            # make a sink with an array to receive visibilities
            ndirs = model_source._nclus
            model_shape = (ndirs, 1, self._expected_nrows, self.nfreq, self.tile.dh.ncorr)
            use_double = self.tile.dh.mb_opts['dtype'] == 'double'
            full_model = np.zeros(model_shape, np.complex128 if use_double else self.tile.dh.ctype)
            print("montblanc dtype is {} ('{}')".format(full_model.dtype, self.tile.dh.mb_opts['dtype']), file=log(2))
            column_snk = MBTiggerSim.ColumnSinkProvider(self.tile.dh, self._freqs.shape, full_model, self._mb_sorted_ind)
            snks = [column_snk]

            for direction in range(ndirs):
                for req_beam in ["beam", "nobeam"]:
                    tigger_source.set_direction(direction, req_beam)
                    tigger_source.set_frequency(self._freqs)
                    column_snk.set_direction(direction)
                    if self._mb_arbeam_src and req_beam == "beam":
                        dir_srcs = srcs + [self._mb_arbeam_src]
                    else:
                        dir_srcs = srcs
                    MBTiggerSim.simulate(dir_srcs,
                                         snks,
                                         polarisation_type=self.tile.dh._poltype,
                                         opts=self.tile.dh.mb_opts)

            # convert back to float, if double was used
            if full_model.dtype != self.tile.dh.ctype:
                full_model = full_model.astype(self.tile.dh.ctype)

            # now associate each cluster in the LSM with an entry in the loaded_models cache
            loaded_models[model_source] = { clus: full_model[i, 0, self._row_identifiers, :, :]
                                                for i, clus in enumerate(tigger_source._cluster_keys)}

            model = loaded_models[model_source][cluster]
            print("  using {}{} for model {} direction {}".format(model_source,
                                "" if not cluster else ("()" if cluster == 'die' else "({})".format(cluster)),
                                imod, idir), file=log(1))

            # release memory asap
            del column_snk, snks

            return model

        def _column_to_cube(self, column, chunk_tdim, chunk_fdim, rows, freqs, dtype, zeroval=0, reqdims=6,
                            allocator=np.empty):
            """
            Converts input data into N-dimensional measurement matrices.

            Args:
                column (np.ndarray):
                    column array from which this will be filled
                chunk_tdim (int):
                    Timeslots per chunk.
                chunk_fdim (int):
                    Frequencies per chunk.
                rows (np.ndarray):
                    Row slice (or set of indices) to be applied to column data in order to extract the required chunk
                freqs (slice):
                    Frequency slice.
                dtype (various):
                    Data type of the resulting measurement matrix.
                zeroval (various, optional):
                    Null value with which to fill missing array elements.
                reqdims (int):
                    Required number of output dimensions.
                allocator (callable):
                    Function to call to allocate empty array. Must have signature (shape,dtype).
                    Default is np.empty.

            Returns:
                np.ndarray:
                    Output cube of with reqdims axes.
            """

            # Start by establishing the possible dimensions and those actually present. Dimensions which
            # are not present are set to one, for flexibility reasons. Output shape is determined by
            # reqdims, which selects dimensions in reverse order from (ndir, nmod, nt, nf, na, na, nc).
            # NOTE: The final dimension will be reshaped into 2x2 blocks outside this function.

            col_ndim = column.ndim

            possible_dims = ["dirs", "mods", "rows", "freqs", "cors"]

            dims = {possible_dims[-i]: column.shape[-i] for i in range(1, col_ndim + 1)}

            dims.setdefault("mods", 1)
            dims.setdefault("dirs", 1)

            out_shape = [dims["dirs"], dims["mods"], chunk_tdim, chunk_fdim, self.nants, self.nants, 2, 2]
            out_shape = out_shape[-reqdims:]

            # this shape has "4" at the end instead of 2,2
            out_shape_4 = out_shape[:-2] + [4]

            # Creates empty N-D array into which the column data can be packed.
            # Place the antenna axes first, for better performance in the kernels

            out_arr0 = allocator(out_shape, dtype)
            out_arr0.fill(zeroval)

            # view onto output array with 4 corr axis rather than 2,2
            out_arr = out_arr0.reshape(out_shape_4)

            # Grabs the relevant time and antenna info.

            achunk = self.antea[rows]
            bchunk = self.anteb[rows]
            tchunk = self.times[rows].copy()
            tchunk -= np.min(tchunk)

            # Creates lists of selections to make subsequent selection from column and out_arr easier.

            corr_slice = slice(None) if self.ncorr == 4 else slice(None, None, 3)

            col_selections = [tuple([dirs, mods, rows, freqs, slice(None)][-col_ndim:])
                              for dirs in range(dims["dirs"]) for mods in range(dims["mods"])]

            cub_selections = [[dirs, mods, tchunk, slice(None), achunk, bchunk, corr_slice][-(reqdims - 1):]
                              for dirs in range(dims["dirs"]) for mods in range(dims["mods"])]

            # The following takes the arbitrarily ordered data from the MS and places it into a N-D
            # data structure (correlation matrix).

            for col_selection, cub_selection in zip(col_selections, cub_selections):

                if self.ncorr == 4:
                    out_arr[tuple(cub_selection)] = colsel = column[col_selection]
                    cub_selection[-3], cub_selection[-2] = cub_selection[-2], cub_selection[-3]
                    if np.iscomplexobj(out_arr):
                        out_arr[tuple(cub_selection)] = colsel.conj()[..., (0, 2, 1, 3)]
                    else:
                        out_arr[tuple(cub_selection)] = colsel[..., (0, 2, 1, 3)]

                elif self.ncorr == 2:
                    out_arr[tuple(cub_selection)] = colsel = column[col_selection]
                    cub_selection[-3], cub_selection[-2] = cub_selection[-2], cub_selection[-3]
                    if np.iscomplexobj(out_arr):
                        out_arr[tuple(cub_selection)] = colsel.conj()
                    else:
                        out_arr[tuple(cub_selection)] = colsel

                elif self.ncorr == 1:
                    out_arr[tuple(cub_selection)] = colsel = column[col_selection][..., (0, 0)]
                    cub_selection[-3], cub_selection[-2] = cub_selection[-2], cub_selection[-3]
                    if np.iscomplexobj(out_arr):
                        out_arr[tuple(cub_selection)] = colsel.conj()
                    else:
                        out_arr[tuple(cub_selection)] = colsel

            return out_arr0

        def _cube_to_column(self, column, in_arr, rows, freqs, flags=False):
            """
            Converts a measurement matrix back into an MS style column.

            Args:
                in_arr (np.ndarray):
                    Input array which is to be made MS friendly.
                rows (np.ndarray):
                    Row indices or slice.
                freqs (slice):
                    Frequency slice.
                flags (bool, optional):
                    If True, input array is a flag cube (i.e. no correlation axes).
            """

            tchunk = self.times[rows].copy()
            tchunk -= tchunk[0]  # is this correct -- does in_array start from beginning of chunk?
            achunk = self.antea[rows]
            bchunk = self.anteb[rows]

            # Flag cube has no correlation axis, so copy it into output column.

            if flags:
                column[rows, freqs, :] = in_arr[tchunk, :, achunk, bchunk, np.newaxis]

            # For other cubes, rehape the last two axes into one (faltten 2x2 correlation block).
            else:
                chunk = in_arr[tchunk, :, achunk, bchunk, :]
                newshape = list(chunk.shape[:-2]) + [chunk.shape[-2] * chunk.shape[-1]]
                chunk = chunk.reshape(newshape)
                if self.ncorr == 4:
                    column[rows, freqs, :] = chunk
                elif self.ncorr == 2:  # 2 corr -- take elements 0,3
                    column[rows, freqs, :] = chunk[..., ::3]
                elif self.ncorr == 1:  # 1 corr -- take element 0
                    column[rows, freqs, :] = chunk[..., :1]


    def __init__(self, handler, chunk):
        """
        Initialises a tile and sets the first row chunk.

        Args:
            handler (:obj:`~cubical.data_handler.ReadModelHandler`):
                Data hander object.
            chunk (:obj:`~cubical.data_handler.RowChunk`):
                Row chunk which is used to initialise the tile.
        """

        self.dh = handler
        self.rowchunks = [chunk]
        self.first_row0 = chunk.rows0[0]
        self.last_row0  = chunk.rows0[-1]
        self._rows_adjusted = False
        self._updated = False
        self.data = None
        self.label = "tile"

    def append(self, chunk):
        """
        Appends a row chunk to a tile.

        Args:
            chunk (:obj:`~cubical.data_handler.RowChunk`):
                Row chunk which will be appended to the assosciated tile.
        """

        self.rowchunks.append(chunk)
        self.first_row0 = min(self.first_row0, chunk.rows0[0])
        self.last_row0  = max(self.last_row0, chunk.rows0[-1])

    def total_tf_chunks(self):
        """Returns total number of time/frequency chunks in the tile, counting the freq chunks in each DDID"""
        return sum([len(self.dh.freqchunks[chunk.ddid]) for chunk in self.rowchunks])

    def merge(self, other):
        """
        Merges another tile into this one.

        Args:
            other (:obj:`~cubical.data_handler.Tile`):
                Tile which will be merged.
        """

        self.rowchunks += other.rowchunks
        self.first_row0 = min(self.first_row0, other.first_row0)
        self.last_row0  = max(self.last_row0, other.last_row0)

    def finalize(self, label=None):
        """
        Creates a list of chunks within the tile that can be iterated over and creates a list of
        chunk labels.

        This also adjusts the row indices of all row chunks so that they become relative to the
        start of the tile.

        Note that this is called in the main process, so all structures set up here will be inherited
        by the I/O and solver workers.
        """
        if label is not None:
            self.label = label

        # row rebin map, relative to start (row0) of tile
        self.rebin_row_map = self.dh.rebin_row_map[self.first_row0:self.last_row0+1]
        first_row, last_row = self.first_row, self.last_row = abs(self.rebin_row_map[0]), abs(self.rebin_row_map[-1])
        num_rows = self.last_row - self.first_row + 1
        # first rebinned row is also 0
        self.rebin_row_map = np.sign(self.rebin_row_map)*(abs(self.rebin_row_map) - first_row)

        self._data_dict_name = "DATA:{}:{}".format(self.first_row0, self.last_row0)

        # Adjust rebinned row indices so that first rebinned row of tile is 0
        if not self._rows_adjusted:
            for rowchunk in self.rowchunks:
                rowchunk.rows -= first_row
            self._rows_adjusted = True

        # Note that first_row0, last_row0 amd rowchunk.rows0 continues to refer to MS rows

        # list of DDIDs in this tile
        self.ddids = np.unique([rowchunk.ddid for rowchunk in self.rowchunks])
        # reverse index: from DDID to its number in the self.ddids list
        self._ddid_reverse = { ddid:num for num, ddid in enumerate(self.ddids) }

        # Create a dict of { chunk_label: rowchunk, chan0, chan1 } for all chunks in this tile.
        self._chunk_dict = OrderedDict()
        self._chunk_indices = {}

        # maps row to chunk. Row 0 is start of tile.
        row_chunknum = np.zeros(num_rows, dtype=int)
        # maps row to DDID. Row 0 is start of tile.
        row_ddid = self.dh.ddid_col[first_row:last_row+1]
        # maps row0 (raw MS rows) to DDID. Row 0 is start of tile.
        row0_ddid = np.zeros(self.last_row0 - self.first_row0 + 1, dtype=int)

        for i, rowchunk in enumerate(self.rowchunks):
            row_chunknum[rowchunk.rows] = i
            row0_ddid[rowchunk.rows0 - self.first_row0] = rowchunk.ddid
            freqchunks = self.dh.freqchunks[rowchunk.ddid]
            for ifreq,chan0 in enumerate(freqchunks):
                key = "D{}T{}F{}".format(rowchunk.ddid, rowchunk.tchunk, ifreq)
                chan1 = freqchunks[ifreq+1] if ifreq<len(freqchunks)-1 else len(self.dh.chanfreqs[rowchunk.ddid])
                self._chunk_dict[key] = rowchunk, chan0, chan1
                self._chunk_indices[key] = rowchunk.tchunk, (rowchunk.ddid, chan0, chan1)

        self.nants = self.dh.nants
        self.ncorr = self.dh.ncorr

        # Set up per-DDID subsets. These will be inherited by workers.

        # A tile consists of 1+ Subsets. One subset if the DDIDs have the same shape, one per DDID if the shape is ragged.
        # A Subset is read from the MS as a single unit (with one getcolnp call)
        # A subset can contain one or more rowchunks.

        # dict: DDID -> subset. For one subset (DDIDs with same shape), None -> subset
        self._subsets = {}

        # if spectral windows are ragged, we generate one Subset object per DDID 
        if self.dh._ddids_unequal:
            subset_row_chunknum = {}
            for ddid in self.ddids:
                #rows = np.array([x[0] for x in ddid_rows[ddid]])
                #rows0 = np.array(sorted(ddid_rows0[ddid]))
                rows = np.where(row_ddid == ddid)[0]
                rows0 = np.where(row0_ddid == ddid)[0] + self.first_row0
                datadict = "{}:D{}".format(self._data_dict_name, ddid)
                self._subsets[ddid] = subset = MSTile.Subset(self, "DDID {}".format(ddid), datadict, rows, rows0)
                # extract chunk numbers for rows of this subset
                subset_row_chunknum[ddid] = np.array(row_chunknum[rows])
            # now for every row chunk, determine which rows from which subset belogn to it
            for i, rowchunk in enumerate(self.rowchunks):
                rowchunk.subset = self._subsets[rowchunk.ddid]
                rowchunk.rows_within_subset = np.where(subset_row_chunknum[rowchunk.ddid] == i)[0]
        # else a single Subset object to read the entire tile at once
        else:
            rows = slice(first_row, last_row + 1)
            rows0 = np.arange(self.first_row0, self.last_row0 + 1)
            self._subsets[None] = subset = MSTile.Subset(self, None, self._data_dict_name, rows, rows0)
            # make a reference from each constituent row chunk to this subset
            for i, rowchunk in enumerate(self.rowchunks):
                rowchunk.subset = subset
                # each row is the actual row within the tile 
                rowchunk.rows_within_subset = np.where(row_chunknum == i)[0]


    def get_chunk_indices(self, key):
        """ Returns chunk indices based on the key value. """

        return self._chunk_indices[key]

    def get_chunk_keys(self):
        """ Returns all chunk keys. """

        return iter(self._chunk_dict.keys())

    def get_chunk_tfs(self, key):
        """
        Returns timestamps and freqs for the given chunk associated with key, as well as two slice
        objects describing its position in the global time/freq space.

        Args:
            key (str):
                The label corresponding to the chunk of interest.

        Returns:
            tuple:
                Unique times, channel frequencies, time axis slice, frequency axis slice (in overall subset of freqs)
        """

        rowchunk, chan0, chan1 = self._chunk_dict[key]
        # lookup ordinal number of this DDID, and convert this to offset in frequencies
        chan_offset = self.dh.ddid_first_chan[rowchunk.ddid]
        return self.dh.uniq_times[rowchunk.timeslice], \
               self.dh.chanfreqs[rowchunk.ddid][chan0:chan1], \
               rowchunk.timeslice, \
               slice(chan_offset + chan0, chan_offset + chan1)

    def fill_bitflags(self, flagbit):
        for subset in self._subsets.values():
            if subset.label is None:
                print("    {}: filling bitflags, MS rows {}~{}".format(self.label, self.first_row0, self.last_row0), file=log(1))
            else:
                print("    {}: filling bitflags, MS rows {}~{}, {} ({} rows)".format(self.label, self.first_row0,
                                                                                          self.last_row0, subset.label,
                                                                                          len(subset.rows0)), file=log(1))
            table_subset = self.dh.data.selectrows(subset.rows0)
            flagcol = table_subset.getcol("FLAG")
            flagrow = table_subset.getcol("FLAG_ROW")
            # flag all four corrs
            num_fl0 = flagcol.sum()
            flagcol[:] = np.logical_or.reduce(flagcol, axis=-1)[:,:,np.newaxis]
            if flagcol.sum() != num_fl0:
                table_subset.putcol("FLAG", flagcol)
                print("    {}: some flags were expanded to all correlations".format(self.label), file=log(1))

            bflagcol = np.zeros(flagcol.shape, np.int32)
            bflagrow = np.zeros(flagrow.shape, np.int32)
            bflagcol[flagcol] = flagbit
            bflagrow[flagrow] = flagbit

            table_subset.putcol("BITFLAG", bflagcol)
            table_subset.putcol("BITFLAG_ROW", bflagrow)

    def load(self, load_model=True):
        """
        Fetches data from MS into tile data shared dict. This is meant to be called in the main
        or I/O process.

        Args:
            load_model (bool, optional):
                If False, omits weights and model visibilities.

        Returns:
            :obj:`~cubical.tools.shared_dict.SharedDict`:
                Shared dictionary containing the MS data relevant to the tile.

        Raises:
            RuntimeError:
                If neither --model-lsm nor --model-column set (see [model] section in
                DefaultParset.cfg).
        """

        # These two variables indicate if the (corrected) data or flags have been updated
        # (Gotcha for shared_dict users! The only truly shared objects are arrays.
        # Thus, we create an array for the two "updated" variables.)

        self._auto_filled_bitflag = False

        # now run over the subsets of the tile set up above. Each subset is a chunk of rows with the same
        # channel shape. If all DDIDs have the same shape, this will be just the one

        for subset in self._subsets.values():
            print(f"{self.label}{' '  + subset.label if subset.label else ''}: reading MS rows {subset.rows0[0]}~{subset.rows0[-1]}", file=log(0, "blue"))

            # create datadict structure for this
            data = shared_dict.create(subset.datadict)
            data.addSubdict('solutions')
            data['updated'] = np.array([False, False, False])

            table_subset = self.dh.data.selectrows(subset.rows0)
            nrows0 = table_subset.nrows()

            original_row_numbers = table_subset.rownumbers(self.data)

            obvis0 = self.dh.fetchslice(self.dh.data_column, subset=table_subset).astype(self.dh.ctype)
            print("  read " + self.dh.data_column, file=log(2))

            uvw0 = table_subset.getcol("UVW")
            print("  read UVW coordinates", file=log(2))

            # read weight columns, if a model is to be read

            if self.dh.has_weights and load_model:
                weights0 = np.zeros([len(self.dh.models)] + list(obvis0.shape), self.dh.wtype)
                wcol_cache = {}
                for imod, (_, weight_columns) in enumerate(self.dh.models):
                    # look for weights to be multiplied in
                    for iwcol, weight_col in enumerate(weight_columns.split("*")):
                        mean_corr = "~ (mean across corrs)" if weight_col.endswith("~") else ""
                        if mean_corr:
                            weight_col = weight_col[:-1]
                        wcol = wcol_cache.get(weight_col)
                        if wcol is None:
                            print("model {} weights {}: reading from {}{}".format(imod, iwcol, weight_col, mean_corr), file=log(2))
                            wcol = table_subset.getcol(str(weight_col))
                            # support two shapes of wcol: either same as data (a-la WEIGHT_SPECTRUM), or missing
                            # a frequency axis (a-la WEIGHT)
                            if wcol.ndim == 3:
                                wcol_cache[weight_col] = wcol = wcol[:, self.dh._channel_slice, self.dh._corr_slice]
                                if wcol.shape != obvis0.shape:
                                    raise RuntimeError("column {} does not match shape of visibility column".format(weight_col))
                            elif tuple(wcol.shape) == (obvis0.shape[0], self.dh.nmscorrs):
                                print("    this weight column does not have a frequency axis: broadcasting", file=log(2))
                                wcol_cache[weight_col] = np.empty_like(obvis0, self.dh.wtype)
                                wcol_cache[weight_col][:] = wcol[:, np.newaxis, self.dh._corr_slice]
                                wcol = wcol_cache[weight_col]
                            else:
                                raise RuntimeError("column {} has an invalid shape {} (expected {})".format(weight_col, wcol.shape, obvis0.shape))
                        else:
                            print("model {} weights {}: reusing {}{}".format(imod, iwcol, weight_col, mean_corr), file=log(2))
                        # take mean weights if specified, else fill off-diagonals
                        if mean_corr:
                            wcol = wcol.mean(-1)[..., np.newaxis]
                        elif self.dh.fill_offdiag_weights and wcol.shape[-1] == 4:
                            wcol[:, :, (1, 2)] = np.sqrt(wcol[: ,: ,0]*wcol[: ,: ,3])[..., np.newaxis]
                        # init weights, if first column, else multiply weights by subsequent column
                        if not iwcol:
                            weights0[imod, ...] = wcol
                        else:
                            weights0[imod, ...] *= wcol
                del wcol_cache
                num_weights = len(self.dh.models)
            else:
                weights0 = np.zeros((len(self.dh.models), 1, 1, 1), self.dh.wtype)
                num_weights = 0

            # The following block of code deals with the various flagging operations and columns. The
            # aim is to correctly populate flag_arr from the various flag sources.

            # Make a flag array. This will contain FL.PRIOR for any points flagged in the MS.

            flag_arr0 = np.zeros(obvis0.shape, dtype=FL.dtype)

            # FLAG/FLAG_ROW only needed if applying them, or auto-filling BITFLAG from them.

            flagcol = flagrow = None
            subset.bflagcol = None
            subset._flagcol_sum = 0
            self.dh.flagcounts["TOTAL"] += flag_arr0.size

            if self.dh._apply_flags:
                flagcol = self.dh.fetchslice("FLAG", subset=table_subset)
                flagcol[:] = np.logical_or.reduce(flagcol, axis=-1)[:,:,np.newaxis]
                flagrow = table_subset.getcol("FLAG_ROW")
                flagcol[flagrow, :, :] = True
                print("  read FLAG/FLAG_ROW", file=log(2))
                # compute stats
                subset._flagcol_sum = flagcol.sum()
                self.dh.flagcounts["FLAG"] += subset._flagcol_sum

                if self.dh._apply_flags:
                    flag_arr0[flagcol] = FL.PRIOR

            # if an active row subset is specified, flag non-active rows as priors. Start as all flagged,
            # the clear the flags
            if self.dh.inactive_rows is not None:
                inactive = self.dh.inactive_rows[original_row_numbers]
            else:
                inactive = np.zeros(nrows0, bool)
            num_inactive = inactive.sum()
            if num_inactive:
                print("  applying a solvable subset deselects {} rows".format(num_inactive), file=log(0))
            # apply baseline selection
            if self.dh.min_baseline or self.dh.max_baseline:
                uv2 = (uvw0[:, 0:2] ** 2).sum(1)
                inactive[uv2 < self.dh.min_baseline ** 2] = True
                if self.dh.max_baseline:
                    inactive[uv2 > self.dh.max_baseline ** 2] = True
                print("  applying solvable baseline cutoff deselects {} rows".format(
                    inactive.sum() - num_inactive), file=log(0))
                num_inactive = inactive.sum()
            if num_inactive:
                print("  {:.2%} visibilities deselected via specificed subset and/or baseline cutoffs".format(num_inactive / float(inactive.size)), file=log(0))
                flag_arr0[inactive] |= FL.SKIPSOL
                self.dh.flagcounts["DESEL"] += num_inactive*flag_arr0[0].size

            # Form up bitflag array, if needed.
            if self.dh._apply_bitflags or self.dh._save_bitflag:
                # If not explicitly re-initializing, try to read column.
                subset.bflagrow = table_subset.getcol("BITFLAG_ROW")
                # If there's an error reading BITFLAG, it must be unfilled. This is a common
                # occurrence so we may as well deal with it. In this case, if auto-fill is set,
                # fill BITFLAG from FLAG/FLAG_ROW.
                subset.bflagcol = self.dh.fetchslice("BITFLAG", subset=table_subset)
                subset.bflagcol[:] = np.bitwise_or.reduce(subset.bflagcol, axis=2)[:,:,np.newaxis]

                print("  read BITFLAG/BITFLAG_ROW", file=log(2))
                # compute stats
                for flagset, bitmask in self.dh.bitflags.items():
                    flagged = subset.bflagcol & bitmask != 0
                    flagged[subset.bflagrow & bitmask != 0, :, :] = True
                    self.dh.flagcounts[flagset] += flagged.sum()

                # apply
                if self.dh._apply_bitflags:
                    flag_arr0[(subset.bflagcol & self.dh._apply_bitflags) != 0] = FL.PRIOR
                    flag_arr0[(subset.bflagrow & self.dh._apply_bitflags) != 0, :, :] = FL.PRIOR

            # if bitflag column is not kept, yet we need to save flags, keep the flag column
            if subset.bflagcol is None and self.dh._save_flags:
                subset._flagcol = flagcol

            flagged = flag_arr0 != 0

            # check for invalid or null-diagonal (diagonal being 0th and last correlation) data
            invalid = ~np.isfinite(obvis0)
            invalid[...,(0,-1)] |= (obvis0[...,(0,-1)]==0)
            invalid &= ~flagged
            ninv = invalid.sum()
            if ninv:
                flagged |= invalid
                flag_arr0[invalid] |= FL.INVALID
                self.dh.flagcounts.setdefault("INVALID", 0)
                self.dh.flagcounts["INVALID"] += ninv
                log.warn("  {:.2%} input visibilities flagged as invalid (0/inf/nan)".format(ninv / float(flagged.size)), "red")

            # check for invalid weights
            if self.dh.has_weights and load_model:
                invalid = (~(np.isfinite(weights0).all(axis=0))) & ~flagged
                ninv = invalid.sum()
                if ninv:
                    flagged |= invalid
                    flag_arr0[invalid] |= FL.INVWGHT
                    self.dh.flagcounts.setdefault("INVWGHT", 0)
                    self.dh.flagcounts["INVWGHT"] += ninv
                    log.warn("  {:.2%} input visibilities flagged due to inf/nan weights".format(
                        ninv / float(flagged.size)), "red")
                wnull = (weights0 == 0).all(axis=0) & ~flagged
                nnull = wnull.sum()
                if nnull:
                    flagged |= wnull
                    flag_arr0[wnull] |= FL.NULLWGHT
                    self.dh.flagcounts.setdefault("NULLWGHT", 0)
                    self.dh.flagcounts["NULLWGHT"] += nnull
                    log.warn("  {:.2%} input visibilities flagged due to null weights".format(
                              nnull / float(flagged.size)), "red")

            nfl = flagged.sum()
            self.dh.flagcounts["IN"] += nfl
            print("  {:.2%} input visibilities flagged and/or deselected".format(nfl / float(flagged.size)), file=log)

            # now rebin arrays if appropriate
            if self.dh.do_freq_rebin or self.dh.do_time_rebin:
                nrows = abs(subset.rebin_row_map[-1])+1
                nchan = subset.rebin_chan_map[-1]+1
                print("  rebinning into {} rows and {} channels".format(nrows, nchan), file=log(0))

                import cubical.kernels
                rebinning = cubical.kernels.import_kernel("rebinning")
                obvis = data.addSharedArray('obvis', [nrows, nchan, self.dh.ncorr], obvis0.dtype)
                rebin_factor = obvis0.size / float(obvis.size)
                flag_arr = data.addSharedArray('flags', obvis.shape, FL.dtype)
                ### no longer filling with -1 -- see flag logic changes in the kernel
                ## flag_arr.fill(-1)
                uvwco = data.addSharedArray('uvwco', [nrows, 3], float)
                # make dummy weight array if not 0
                if num_weights:
                    weights = data.addSharedArray('weigh', [num_weights] + list(obvis.shape), self.dh.wtype)
                else:
                    weights = np.zeros((1, 1, 1, 1), self.dh.wtype)

                rebinning.rebin_vis(obvis, obvis0, uvwco, uvw0,
                                      flag_arr, flag_arr0,
                                      weights, weights0, num_weights,
                                      subset.rebin_row_map, subset.rebin_chan_map)
#                import pdb; pdb.set_trace()
                del obvis0, uvw0
                # we'll need flag_arr0 and weights0 for load_models below so don't delete
                flagged = flag_arr != 0
            # else copy arrays to shm directly
            else:
                rebin_factor = 1
                # still need to adjust conjugate rows
                obvis0[subset.rebin_row_map<0] = obvis0[subset.rebin_row_map<0].conjugate()
                nrows = nrows0
                data['obvis'] = obvis0
                obvis = data['obvis']   # can't string assigmnent together: x = dict['key'] = y does not guarantee that x is dict['key']!!!
                data['flags'] = flag_arr0
                flag_arr = data['flags']
                data['uvwco'] = uvw0
                uvwco = data['uvwco']
                if num_weights:
                    data['weigh'] = weights0
                del obvis0, uvw0, weights0

            # normalize by amplitude, if needed
#            if self.dh.do_normalize_data:
#                dataabs = np.abs(obvis)
#                with np.errstate(divide='ignore'):
#                    obvis /= dataabs
#                    obvis[dataabs==0] = 0
#                if 'weigh' in data:
#                    data['weigh'] *= dataabs[np.newaxis, ...]
#                else:
#                    data['weigh'] = dataabs.reshape([1]+list(dataabs.shape))

            # compute PA rotation angles, if needed
            if self.dh.parallactic_machine is not None:
                angles = self.dh.parallactic_machine.rotation_angles(subset.time_col)
                data.addSharedArray('pa', angles.shape, np.float64)
                data['pa'][:] = angles

            # The following either reads model visibilities from the measurement set, or uses an lsm
            # and Montblanc to simulate them. Data may need to be massaged to be compatible with
            # Montblanc's strict requirements.

            if load_model:
                model_shape = [len(self.dh.model_directions), len(self.dh.models)] + list(obvis.shape)
                loaded_models = {}
                movis = data.addSharedArray('movis', model_shape, self.dh.ctype)

                for imod, (dirmodels, _) in enumerate(self.dh.models):
                    # populate directions of this model
                    for idir, dirname in enumerate(self.dh.model_directions):
                        if dirname in dirmodels:
                            # loop over additive components
                            for model_source, cluster, subtract in dirmodels[dirname]:
                                subtract_str = " (-)" if subtract else ""
                                # see if data for this model is already loaded
                                if model_source in loaded_models:
                                    print("  reusing {}{} for model {} direction {}{}".format(model_source,
                                                                                                      "" if not cluster else (
                                                                                                          "()" if cluster == 'die' else "({})".format(
                                                                                                              cluster)),
                                                                                                      imod, idir, subtract_str), file=log(2))
                                    model = loaded_models[model_source][cluster]
                                # cluster of None signifies that this is a visibility column
                                elif cluster is None:

                                    if model_source == 1:
                                        print("  using 1.+0j for model {} direction {}{}".format(model_source,
                                                                                                         imod, idir, subtract_str), file=log(2))
                                        model = np.ones_like(obvis)
                                    else:
                                        print("  reading {} for model {} direction {}{}".format(model_source, imod,
                                                                                                        idir, subtract_str), file=log(2))
                                        model0 = self.dh.fetchslice(model_source, subset=table_subset)
                                        # sanity check (I've seen nulls coming out of wsclean...)
                                        invmodel = (~np.isfinite(model0))
                                        invmodel[..., (0, -1)] |= (model0[..., (0, -1)] == 0)
                                        invmodel &= (flag_arr0==0)
                                        num_inv = invmodel.sum()
                                        if num_inv:
                                            print("  {} ({:.2%}) model visibilities flagged as 0/inf/nan".format(
                                                num_inv, num_inv / float(invmodel.size)), file=log(0,"red"))
                                            flag_arr0[invmodel] |= FL.INVMODEL
                                        # now rebin (or conjugate)
                                        if self.dh.do_freq_rebin or self.dh.do_time_rebin:
                                            model = np.zeros_like(obvis)
                                            rebinning.rebin_model(model, model0, flag_arr0,
                                                                  subset.rebin_row_map, subset.rebin_chan_map)
#                                            import pdb; pdb.set_trace()
                                        else:
                                            model0[subset.rebin_row_map < 0] = model0[subset.rebin_row_map < 0].conjugate()
                                            model = model0
                                        model0 = None
                                        # apply rotation (*after* rebinning: subset.time_col is rebinned version!)
                                        if self.dh.parallactic_machine is not None:
                                            subset._angles = self.dh.parallactic_machine.rotation_angles(subset.time_col)
                                            model = self.dh.parallactic_machine.rotate(subset.time_col, model,
                                                                                        subset.antea, subset.anteb,
                                                                                        angles=subset._angles)
                                    loaded_models.setdefault(model_source, {})[None] = model
                                # else evaluate a Tigger model with Montblanc
                                elif TiggerSourceProvider is not None and isinstance(model_source, TiggerSourceProvider):
                                    model = subset.load_montblanc_models(uvwco, loaded_models, model_source, cluster, imod, idir)
                                    # montblanc applies the PA rotation matrix internally
                                elif DicoSourceProvider is not None and isinstance(model_source, DicoSourceProvider):

                                    if self.dh.ncorr == 4:
                                        model_type="cplx2x2"
                                    elif self.dh.ncorr == 2:
                                        model_type="cplxdiag"
                                    elif self.dh.ncorr == 1:
                                        model_type="cplxscalar"
                                    else:
                                        raise RuntimeError("Visibilities have correlations other than 4, 2 or 1. At present this is not supported")
                                    model = subset.load_ddfacet_models(uvwco, loaded_models, model_source, cluster, imod, idir, model_type)
                                    if self.dh.parallactic_machine is not None:
                                            #subset._angles = self.dh.parallactic_machine.rotation_angles(subset.time_col)
                                            subset._angles = data['pa'][:]
                                            model = self.dh.parallactic_machine.rotate(subset.time_col, model,
                                                                                       subset.antea, subset.anteb,
                                                                                       angles=subset._angles)
                                else:
                                    raise TypeError("Unknown cluster of type {0:s}".format(type(model_source)))
                                # finally, add model in at correct slot
                                if subtract:
                                    movis[idir, imod, ...] -= model
                                else:
                                    movis[idir, imod, ...] += model
                                del model

                # round to 1e-16 to avoid underflows (casa setjy, for example, can yield visibilities with a tiny
                # imaginary part, which cause underflows when squared)
                movis.round(16, out=movis)

                # check for a null model (all directions)
                invmodel = (~np.isfinite(movis)).any(axis=(0,1))
                invmodel[...,(0,-1)] |= (movis[...,(0,-1)]==0).all(axis=(0,1))
                invmodel &= ~flagged

                ninv = invmodel.sum()
                if ninv:
#                    import pdb; pdb.set_trace()
                    flag_arr[invmodel] |= FL.INVMODEL
                    self.dh.flagcounts.setdefault("INVMODEL", 0)
                    self.dh.flagcounts["INVMODEL"] += ninv*rebin_factor
                    print("  {} ({:.2%}) visibilities flagged due to 0/inf/nan model".format(
                        ninv, ninv / float(flagged.size)), file=log(0, "red"))

                # release memory (gc.collect() particularly important), as model visibilities are *THE* major user (especially
                # in the DD case)
                del loaded_models, flag_arr0
                import gc
                gc.collect()

            data.addSharedArray('covis', data['obvis'].shape, self.dh.ctype)
#            import pdb; pdb.set_trace()

            # Create a placeholder if using the Robust solver with save weights activated
            if self.dh.output_weight_column is not None:
                data.addSharedArray('outweights', data['obvis'].shape, self.dh.wtype)

    def get_chunk_cubes(self, key, ctype=np.complex128, allocator=np.empty, flag_allocator=np.empty):
        """
        Produces the CubiCal data cubes corresponding to the specified key.

        Args:
            key (str):
                The label corresponding to the chunk of interest.
            ctype (type):
                Data type of complex arrays.
            allocator (callable):
                Function called to allocate array. Signature (shape,dtype)
            flag_allocator (callable):
                Function called to allocate flag-like arrays. Signature (shape,dtype)


        Returns:
            tuple:
                The data, model, flags and weights cubes for the given chunk key.
                Shapes are as follows:

                - data (np.ndarray):    [n_tim, n_fre, n_ant, n_ant, 2, 2]
                - model (np.ndarray):   [n_dir, n_mod, n_tim, n_fre, n_ant, n_ant, 2, 2]
                - flags (np.ndarray):   [n_tim, n_fre, n_ant, n_ant]
                - weights (np.ndarray): [n_mod, n_tim, n_fre, n_ant, n_ant, 2, 2] or None for no weighting

                n_mod refers to number of models simultaneously fitted.
        """
        rowchunk, freq0, freq1 = self._chunk_dict[key]
        subset, row_index = rowchunk.subset, rowchunk.rows_within_subset

        data = shared_dict.attach(subset.datadict)

        t_dim = self.dh.chunk_ntimes[rowchunk.tchunk]
        f_dim = freq1 - freq0
        freq_slice = slice(freq0, freq1)

        log(1).print(f"getting cubes for subset {subset.label}: {len(row_index)} rows, first is {row_index[0]}, last is {row_index[-1]}")
        flags_2x2 = subset._column_to_cube(data['flags'], t_dim, f_dim, row_index, freq_slice,
                                         FL.dtype, FL.MISSING, allocator=allocator)
        flags = flag_allocator(flags_2x2.shape[:-2], flags_2x2.dtype)
        if self.ncorr == 4:
            np.bitwise_or.reduce(flags_2x2, axis=(-1, -2), out=flags)
        else:
            flags[:] = flags_2x2[..., 0, 0]
            flags |= flags_2x2[..., 1, 1]

        # These points correspond to the autocorrelations - they should always
        # be flagged for the purpose of calibration.
        assert flags.ndim == 4 and flags.shape[2] == flags.shape[3]
        nants = flags.shape[3]
        flags[..., range(nants), range(nants)] |= FL.SKIPSOL

        obs_arr = subset._column_to_cube(data['obvis'], t_dim, f_dim, row_index, freq_slice, ctype,
                                       reqdims=6, allocator=allocator)
        if 'movis' in data:
            ### APPLY ROTATION HERE
            # @ oms this is wrong when mixing and matching predictors and very likely leeds
            # to doubly applying the PA model rotation.
            # if self.dh.rotate_model:
            #     model = data['movis']
            #     angles = data['pa'][rows]
            #     for idir in range(model.shape[0]):
            #         for imod in range(model.shape[1]):
            #             submod = model[idir, imod, rows, freq_slice]
            #             submod[:] = self.dh.parallactic_machine.rotate(subset.time_col[rows], submod,
            #                                                subset.antea[rows], subset.anteb[rows],
            #                                                angles=angles)

            mod_arr = subset._column_to_cube(data['movis'], t_dim, f_dim, row_index, freq_slice, ctype,
                                           reqdims=8, allocator=allocator)
            # flag invalid model visibilities
            flags[(~np.isfinite(mod_arr[0, 0, ...])).any(axis=(-2, -1))] |= FL.INVALID
        else:
            mod_arr = None

        if 'weigh' in data:
            wgt_arr = subset._column_to_cube(data['weigh'], t_dim, f_dim, row_index, freq_slice, self.dh.wtype,
                                           allocator=allocator)
            # wgt_arr = flag_allocator(wgt_2x2.shape[:-2], wgt_2x2.dtype)
            # np.mean(wgt_2x2, axis=(-1, -2), out=wgt_arr)
            # #            wgt_arr = np.sqrt(wgt_2x2.sum(axis=(-1,-2)))    # this is wrong
            wgt_arr[flags!=0, :, :] = 0  # autocorrs will get weight zero.
            # wgt_arr = wgt_arr.reshape([1, t_dim, f_dim, nants, nants])
        else:
            wgt_arr = None

        # # zero flagged entries in data and model. NB: this is now done in the solver instead
        # obs_arr[flagged, :, :] = 0
        # if mod_arr is not None:
        #     mod_arr[0, 0, flagged, :, :] = 0

        return obs_arr, mod_arr, flags, wgt_arr

    def set_chunk_cubes(self, cube, flag_cube, weight_cube, key, column='covis'):
        """
        Copies a visibility cube, and an optional flag cube, back to tile column.

        Args:
            cube (np.ndarray):
                Cube containing visibilities.
            flag_cube (np.ndarray):
                Cube containing flags.
            weight_cube (np.ndarray):
                Cube containing weights
            key (str):
                The label corresponding to the chunk of interest.
            column (str, optional):
                The column to which the cube must be copied.
        """
        rowchunk, freq0, freq1 = self._chunk_dict[key]
        subset, row_index = rowchunk.subset, rowchunk.rows_within_subset
        data = shared_dict.attach(subset.datadict)
        freq_slice = slice(freq0, freq1)

        if cube is not None:
            data['updated'][0] = True
            subset._cube_to_column(data[column], cube, row_index, freq_slice)

            ### APPLY DEROTATION HERE
            if self.dh.derotate_output:
                data[column][row_index, freq_slice] = self.dh.parallactic_machine.derotate(subset.time_col[row_index],
                                                               data[column][row_index, freq_slice],
                                                               subset.antea[row_index], subset.anteb[row_index],
                                                               angles=data['pa'][row_index])
        if flag_cube is not None:
            data['updated'][1] = True
            subset._cube_to_column(data['flags'], flag_cube, row_index, freq_slice, flags=True)
        if weight_cube is not None:
           data['updated'][2] = True
           subset._cube_to_column(data['outweights'], weight_cube, row_index, freq_slice)

    def _get_chunk_datadict(self, key):
        rowchunk, _, _ = self._chunk_dict[key]
        return shared_dict.attach(rowchunk.subset.datadict)

    def create_solutions_chunk_dict(self, key):
        """
        Creates a shared dict for the given chunk in which to store gain solutions.

        Args:
            key (str):
                The label corresponding to the chunk of interest.

        Returns:
            :obj:`~cubical.tools.shared_dict.SharedDict`:
                Shared dictionary containing gain solutions.
        """
        data = self._get_chunk_datadict(key)
        sd = data['solutions'].addSubdict(key)

        return sd

    def iterate_solution_chunks(self):
        """
        Iterates over per-chunk solution dictionaries.

        Yields:
            tuple:
                A gain subdictionary and the time and frequency slices to which it corresponds:

                - Subdictionary (:obj:`~cubical.tools.shared_dict.SharedDict`))
                - Time slice (slice)
                - Frequency slice (slice)
        """
        for subset in self._subsets.values():
            data = shared_dict.attach(subset.datadict)
            soldict = data['solutions']
            for key in soldict.keys():
                yield soldict[key]

    def save(self, final=False, only_save=["output", "model", "weight", "flag", "bitflag"]):
        """
        Saves 'corrected' column, and any updated flags, back to MS.

        Args:
            final (bool, optional):
                If True, tells the MS handler that this is the final tile.
            only_save (list, optional):
                Only saves "output", "model", "weight" "flag" or "bitflag". This is useful for
                predicting
        """
        # insert columns first, if needed, and reopen MS
        added = False

        for subset in self._subsets.values():
            print(f"{self.label}{' '  + subset.label if subset.label else ''}: saving MS rows {subset.rows0[0]}~{subset.rows0[-1]}", file=log(0, "blue"))
            data = shared_dict.attach(subset.datadict)

            # insert output columns, if needed, and reopen MS if they were actually added
            if not added:
                if self.dh.output_column and data['updated'][0]:
                    added = self.dh._add_column(self.dh.output_column)
                if self.dh.output_model_column and 'movis' in data:
                    added = self.dh._add_column(self.dh.output_model_column) or added
                if self.dh.output_weight_column and data['updated'][2]:
                    added = self.dh._add_column(self.dh.output_weight_column, like_type='float') or added
                if added:
                    self.dh.reopen()
                added = True

            table_subset = self.dh.data.selectrows(subset.rows0)

            if self.dh.output_column and data['updated'][0]:
                if ("output" in only_save):
                    covis = data['covis']
                    # if self.dh.parallactic_machine is not None:
                    #     covis = self.dh.parallactic_machine.derotate(subset.time_col, covis, subset.antea, subset.anteb,
                    #                                                  angles=subset._angles)
                    covis = subset.upsample(covis)
                    print("  writing {} column".format(self.dh.output_column), file=log)
                    self.dh.putslice(self.dh.output_column, covis, subset=table_subset)
                else:
                    print("  skipping {} column per user request".format(self.dh.output_column), file=log)
            subset._angles = None

            if self.dh.output_model_column and 'movis' in data:
                if ("model" in only_save):
                    # take first mode, and sum over directions if needed
                    model = data['movis'][:, 0]
                    if model.shape[0] == 1:
                        model = model.reshape(model.shape[1:])
                    else:
                        model = model.sum(axis=0)
                    model = subset.upsample(model)
                    print("  writing {} column".format(self.dh.output_model_column), file=log)
                    self.dh.putslice(self.dh.output_model_column, model, subset=table_subset)
                else:
                    print("  skipping {} column per user request".format(self.dh.output_model_column), file=log)
            #writing outputs weights if any
            if self.dh.output_weight_column and data['updated'][2]:
                if ("weight" in only_save):
                    outweights = subset.upsample(data['outweights'])
                    print("  writing {} weight column".format(self.dh.output_weight_column), file=log)
                    self.dh.putslice(self.dh.output_weight_column, outweights, subset=table_subset)
                else:
                    print("  skipping {} weight column per user request".format(self.dh.output_weight_column), file=log)
            # write flags if (a) solver has generated flags, and we're saving them, (b) always, if auto-filling BITFLAG column
            #
            flag_col = None     # if not None, FLAG/FLAG_ROW will be saved
            bflag_col = None    # if not None, BITFLAG/BITFLAG_ROW will be saved

            if data['updated'][1]:
                # count new flags
                newflags = subset.upsample(data['flags'] & ~(FL.PRIOR | FL.SKIPSOL) != 0)
                nfl = newflags.sum()
                # add to stats
                self.dh.flagcounts['NEW'] += nfl
                ratio = nfl / float(newflags.size)

                # write to BITFLAG, if asked to
                if self.dh._save_bitflag:
                    # clear bitflag column first
                    subset.bflagcol &= ~self.dh._save_bitflag
                    subset.bflagcol[newflags] |= self.dh._save_bitflag
                    bflag_col = True
                    if self.dh._save_flags:
                        print("  {:.2%} visibilities flagged by solver: saving to BITFLAG and FLAG columns".format(ratio), file=log)
                        flag_col = subset.bflagcol != 0
                    else:
                        print("  {:.2%} visibilities flagged by solver: saving to BITFLAG column only".format(ratio), file=log)

                # else write to FLAG/FLAG_ROW only, if asked to
                elif self.dh._save_flags:
                    print("  {:.2%} visibilities flagged by solver: saving to FLAG column".format(ratio), file=log)
                    subset._flagcol[newflags] = True
                    flag_col = subset._flagcol

                # else just message
                else:
                    print("  {:.2%} visibilities flagged by solver, but we're not saving flags".format(ratio), file=log)
            else:
                 print("  no solver flags were generated", file=log)
                 # bitflag still needs to be cleared though, if we're writing to it
                 if self.dh._save_bitflag:
                     subset.bflagcol &= ~self.dh._save_bitflag
                     bflag_col = True
                     if self.dh._save_flags:
                         print("  no visibilities flagged by solver: saving to BITFLAG and FLAG columns", file=log)
                         flag_col = subset.bflagcol != 0
                     else:
                         print("  no visibilities flagged by solver: saving to BITFLAG column only", file=log)

            if self.dh._save_flags_apply:
                prior_flags = subset.upsample((data['flags'] & FL.PRIOR) != 0)
                if flag_col is None:
                    flag_col = prior_flags
                else:
                    flag_col |= prior_flags
                ratio = prior_flags.sum() / float(prior_flags.size)
                print("  also transferring {:.2%} input flags (--flags-save-legacy apply)".format(ratio), file=log)
                flag_col[:] = np.logical_or.reduce(axis=-2)[:,:,np.newaxis]

            # now figure out what to write
            # this is set if BITFLAG/BITFLAG_ROW is to be written out
            if bflag_col is not None:
                if ("bitflag" in only_save):
                    subset.bflagcol[:] = np.bitwise_or.reduce(subset.bflagcol, axis=2)[:,:,np.newaxis]
                    self.dh.putslice("BITFLAG", subset.bflagcol, subset=table_subset)
                    totflags = (subset.bflagcol != 0).sum()
                    self.dh.flagcounts['OUT'] += totflags
                    print("  updated BITFLAG column ({:.2%} visibilities flagged)".format(totflags / float(subset.bflagcol.size)), file=log)
                    subset.bflagrow = np.bitwise_and.reduce(subset.bflagcol, axis=(-1, -2))
                    table_subset.putcol("BITFLAG_ROW", subset.bflagrow)
                    print("  updated BITFLAG_ROW column ({:.2%} rows flagged)".format(
                        (subset.bflagrow!=0).sum()/float(subset.bflagrow.size)), file=log)
                else:
                    totflags = (subset.bflagcol != 0).sum()
                    print("  skipping BITFLAG column per user request ({:.2%} visibilities would have been flagged otherwise)".format(totflags / float(subset.bflagcol.size)), file=log)
                    print("  skipping BITFLAG column per user request ({:.2%} visibilities would have been flagged otherwise)".format(totflags / float(subset.bflagcol.size)), file=log)

            #prevents memory leak by clearing
            subset.bflagcol = subset.bflagrow = None

            # this is set if FLAG/FLAG_ROW is to be written out
            if flag_col is not None:
                if ("flag" in only_save):
                    self.dh.putslice("FLAG", flag_col, subset=table_subset)
                    totflags = flag_col.sum()
                    if bflag_col is None:                  # only count if not counted above
                        self.dh.flagcounts['OUT'] += totflags
                    print("  updated FLAG column ({:.2%} visibilities flagged)".format(totflags / float(flag_col.size)), file=log)
                    flag_row = flag_col.all(axis=(-1, -2))
                    table_subset.putcol("FLAG_ROW", flag_row)
                    print("  updated FLAG_ROW column ({:.2%} rows flagged)".format(flag_row.sum() / float(flag_row.size)), file=log)
                else:
                    totflags = flag_col.sum()
                    flag_row = flag_col.all(axis=(-1, -2))
                    print("  skipping FLAG column per user request ({:.2%} visibilities would have been flagged otherwise)".format(totflags / float(flag_col.size)), file=log)
                    print("  skipping FLAG_ROW column per user request ({:.2%} rows would have been flagged otherwise)".format(flag_row.sum() / float(flag_row.size)), file=log)

            # Close the Montblanc providers, if any
            for provider in (subset._mb_measet_src, subset._mb_cached_ms_src):
                if provider is not None:
                    provider.close()

            if final:
                self.dh.finalize()
                self.dh.unlock()
                # close the singleton beam provider
                if subset._mb_arbeam_src is not None:
                    subset._mb_arbeam_src.close()

    def release(self, final=False):
        """ Releases the shared memory data dicts. """

        data = shared_dict.attach(self._data_dict_name)
        data.delete()
        for subset in self._subsets.values():
            if subset.label is not None:
                data = shared_dict.attach(subset.datadict)
                data.delete()
        # clean up DDF degridder stuff, but only if it was already imported
        if final and 'cubical.degridder.DDFacetSim' in sys.modules:
            try:
                from cubical.degridder.DDFacetSim import cleanup_degridders
            except ImportError:
                for line in traceback.format_exc().splitlines(keepends=False):
                    log.info(line)
                log.warn("Error importing cubical.degridder.DDFacetSim.cleanup_degridders, see above")
                def cleanup_degridders(): pass
            cleanup_degridders()






# CubiCal: a radio interferometric calibration suite
# (c) 2017 Rhodes University & Jonathan S. Kenyon
# http://github.com/ratt-ru/CubiCal
# This code is distributed under the terms of GPLv2, see LICENSE.md for details

import numpy as np
import sys

def uniquify(values):
    """
    Helper function. Given a vector of values (e.g. times), finds the set of unique values,
    and computes a vector of indices (e.g. timeslots) of the same size, as well as rmap: a map from the
    unique values to their index numbers.

    Returns tuple of indices, unique_values, rmap
    """
    uniq = np.unique(values) #np.array(sorted(set(values)))
    rmap = {x: i for i, x in enumerate(uniq)}
    # apply this map to the time column to construct a timestamp column
    indices  = np.fromiter(list(map(rmap.__getitem__, values)), int)
    return indices, uniq, rmap

# Try to import montblanc: if not successful, remember error for later.
def import_montblanc():
    """
    Tries to import montblanc. Returns tuple of montblanc_module, None on success, or
    None, exc_info on error
    """
    try:
        import montblanc
        # all of these potentially fall over if Montblanc is the wrong version or something, so moving them here
        # for now
        from .MBTiggerSim import simulate, MSSourceProvider, ColumnSinkProvider
        from .TiggerSourceProvider import TiggerSourceProvider
        from montblanc.impl.rime.tensorflow.sources import CachedSourceProvider, FitsBeamSourceProvider
        return montblanc, None
    except:
        return None, sys.exc_info()

# Try to import ddfacet: if not successful, remember error for later.
def import_ddfacet():
    """
    Tries to import ddfacet. Returns tuple of ddfacet_module, None on success, or
    None, exc_info on error
    """
    try:
        import DDFacet
        return DDFacet, None
    except:
        return None, sys.exc_info()

class Metadata(object):
    """This class holds metadata from an MS"""
    def __init__(self):
        self.num_corrs    = 0
        self.num_antennas = 0
        self.num_baselines = 0
        self.antenna_name = []              # p -> antenna name
        self.antenna_index = {}             # antenna name -> p
        self.baseline_name = {}             # p,q -> baseline name
        self.baseline_length = {}           # p,q -> baseline length
        self.feeds = "xy"                   # "xy" or "rl"
        self.ra0 = 0
        self.dec0 = 0
